#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import copy
import logging
from typing import Any, Dict, List, Optional
from uuid import UUID

from fastapi.encoders import jsonable_encoder
from sqlalchemy import and_, func
from sqlalchemy.orm import Session, contains_eager, joinedload

from argilla.server.contexts import accounts
from argilla.server.enums import ResponseStatusFilter
from argilla.server.models import (
    Dataset,
    DatasetStatus,
    Field,
    Question,
    Record,
    Response,
    ResponseStatus,
    ResponseValue,
)
from argilla.server.schemas.v1.datasets import (
    DatasetCreate,
    FieldCreate,
    QuestionCreate,
    RecordInclude,
    RecordsCreate,
)
from argilla.server.schemas.v1.records import ResponseCreate
from argilla.server.schemas.v1.responses import ResponseUpdate
from argilla.server.search_engine import SearchEngine
from argilla.server.security.model import User

LIST_RECORDS_LIMIT = 20

_LOGGER = logging.getLogger("argilla.server")


def get_dataset_by_id(db: Session, dataset_id: UUID):
    return db.get(Dataset, dataset_id)


def get_dataset_by_name_and_workspace_id(db: Session, name: str, workspace_id: UUID):
    return db.query(Dataset).filter_by(name=name, workspace_id=workspace_id).first()


def list_datasets(db: Session):
    return db.query(Dataset).order_by(Dataset.inserted_at.asc()).all()


def create_dataset(db: Session, dataset_create: DatasetCreate):
    dataset = Dataset(
        name=dataset_create.name,
        guidelines=dataset_create.guidelines,
        workspace_id=dataset_create.workspace_id,
    )

    db.add(dataset)
    db.commit()
    db.refresh(dataset)

    return dataset


async def publish_dataset(db: Session, search_engine: SearchEngine, dataset: Dataset):
    if dataset.is_ready:
        raise ValueError("Dataset is already published")

    if _count_fields_by_dataset_id(db, dataset.id) == 0:
        raise ValueError("Dataset cannot be published without fields")

    if _count_questions_by_dataset_id(db, dataset.id) == 0:
        raise ValueError("Dataset cannot be published without questions")

    dataset.status = DatasetStatus.ready
    db.commit()
    db.expire(dataset)

    try:
        await search_engine.create_index(dataset)
    except:
        # TODO: Improve this action using some rollback mechanism
        dataset.status = DatasetStatus.draft
        db.commit()
        db.expire(dataset)
        raise

    return dataset


async def delete_dataset(db: Session, search_engine: SearchEngine, dataset: Dataset):
    db.delete(dataset)
    await search_engine.delete_index(dataset)

    db.commit()

    return dataset


def get_field_by_id(db: Session, field_id: UUID):
    return db.get(Field, field_id)


def get_field_by_name_and_dataset_id(db: Session, name: str, dataset_id: UUID):
    return db.query(Field).filter_by(name=name, dataset_id=dataset_id).first()


def create_field(db: Session, dataset: Dataset, field_create: FieldCreate):
    if dataset.is_ready:
        raise ValueError("Field cannot be created for a published dataset")

    field = Field(
        name=field_create.name,
        title=field_create.title,
        required=field_create.required,
        settings=field_create.settings.dict(),
        dataset_id=dataset.id,
    )

    db.add(field)
    db.commit()
    db.refresh(field)

    return field


def delete_field(db: Session, field: Field):
    if field.dataset.is_ready:
        raise ValueError("Fields cannot be deleted for a published dataset")

    db.delete(field)
    db.commit()

    return field


def get_question_by_id(db: Session, question_id: UUID):
    return db.get(Question, question_id)


def get_question_by_name_and_dataset_id(db: Session, name: str, dataset_id: UUID):
    return db.query(Question).filter_by(name=name, dataset_id=dataset_id).first()


def create_question(db: Session, dataset: Dataset, question_create: QuestionCreate):
    if dataset.is_ready:
        raise ValueError("Question cannot be created for a published dataset")

    question = Question(
        name=question_create.name,
        title=question_create.title,
        description=question_create.description,
        required=question_create.required,
        settings=question_create.settings.dict(),
        dataset_id=dataset.id,
    )

    db.add(question)
    db.commit()
    db.refresh(question)

    return question


def delete_question(db: Session, question: Question):
    if question.dataset.is_ready:
        raise ValueError("Questions cannot be deleted for a published dataset")

    db.delete(question)
    db.commit()

    return question


def get_record_by_id(db: Session, record_id: UUID):
    return db.get(Record, record_id)


def get_records_by_ids(
    db: Session,
    dataset_id: UUID,
    record_ids: List[UUID],
    include: List[RecordInclude] = [],
    user_id: Optional[UUID] = None,
) -> List[Record]:
    query = db.query(Record).filter(Record.dataset_id == dataset_id, Record.id.in_(record_ids))
    if RecordInclude.responses in include:
        if user_id:
            query = query.outerjoin(
                Response, and_(Response.record_id == Record.id, Response.user_id == user_id)
            ).options(contains_eager(Record.responses))
        else:
            query = query.options(joinedload(Record.responses))
    return query.all()


def list_records_by_dataset_id(
    db: Session,
    dataset_id: UUID,
    include: List[RecordInclude] = [],
    offset: int = 0,
    limit: int = LIST_RECORDS_LIMIT,
):
    query = db.query(Record)

    if RecordInclude.responses in include:
        query = query.options(joinedload(Record.responses))

    return (
        query.filter(Record.dataset_id == dataset_id)
        .order_by(Record.inserted_at.asc())
        .offset(offset)
        .limit(limit)
        .all()
    )


def list_records_by_dataset_id_and_user_id(
    db: Session,
    dataset_id: UUID,
    user_id: UUID,
    include: List[RecordInclude] = [],
    response_status: Optional[ResponseStatusFilter] = None,
    offset: int = 0,
    limit: int = LIST_RECORDS_LIMIT,
):
    query = db.query(Record).filter(Record.dataset_id == dataset_id)

    if response_status == ResponseStatusFilter.missing:
        query = (
            query.outerjoin(
                Response,
                and_(Response.record_id == Record.id, Response.user_id == user_id),
            )
            .filter(Response.status == None)
            .options(contains_eager(Record.responses))
        )
    else:
        if response_status:
            query = query.join(
                Response,
                and_(
                    Response.record_id == Record.id,
                    Response.user_id == user_id,
                    Response.status == ResponseStatus(response_status),
                ),
            ).options(contains_eager(Record.responses))
        elif RecordInclude.responses in include:
            query = query.outerjoin(
                Response,
                and_(Response.record_id == Record.id, Response.user_id == user_id),
            ).options(contains_eager(Record.responses))

    return query.order_by(Record.inserted_at.asc()).offset(offset).limit(limit).all()


def count_records_by_dataset_id(db: Session, dataset_id: UUID):
    return db.query(func.count(Record.id)).filter_by(dataset_id=dataset_id).scalar()


async def create_records(
    db: Session,
    search_engine: SearchEngine,
    dataset: Dataset,
    records_create: RecordsCreate,
):
    if not dataset.is_ready:
        raise ValueError("Records cannot be created for a non published dataset")

    records = []
    for record_create in records_create.items:
        validate_record_fields(dataset, fields=record_create.fields)

        record = Record(
            fields=record_create.fields,
            metadata_=record_create.metadata,
            external_id=record_create.external_id,
            dataset_id=dataset.id,
        )

        if record_create.responses:
            for response in record_create.responses:
                if not accounts.get_user_by_id(db, response.user_id):
                    raise ValueError(f"Provided user_id: {response.user_id!r} is not a valid user id")

                validate_response_values(dataset, values=response.values, status=response.status)

                record.responses.append(
                    Response(values=jsonable_encoder(response.values), status=response.status, user_id=response.user_id)
                )

        records.append(record)

    try:
        db.add_all(records)

        db.commit()
        db.expire_all()

        await search_engine.add_records(dataset, records)
    except:
        # TODO: Improve this action using some rollback mechanism
        for record in records:
            db.delete(record)
        db.commit()
        raise


def get_response_by_id(db: Session, response_id: UUID):
    return db.get(Response, response_id)


def get_response_by_record_id_and_user_id(db: Session, record_id: UUID, user_id: UUID):
    return db.query(Response).filter_by(record_id=record_id, user_id=user_id).first()


def list_responses_by_record_id(db: Session, record_id: UUID):
    return db.query(Response).filter_by(record_id=record_id).order_by(Response.inserted_at.asc()).all()


def count_responses_by_dataset_id_and_user_id(
    db: Session, dataset_id: UUID, user_id: UUID, response_status: Optional[ResponseStatus] = None
) -> int:
    expressions = [Response.user_id == user_id]
    if response_status:
        expressions.append(Response.status == response_status)

    return (
        db.query(func.count(Response.id))
        .join(Record, and_(Record.id == Response.record_id, Record.dataset_id == dataset_id))
        .filter(*expressions)
        .scalar()
    )


def count_records_with_missing_responses_by_dataset_id_and_user_id(db: Session, dataset_id: UUID, user_id: UUID):
    return (
        db.query(Record.id)
        .outerjoin(
            Response,
            and_(
                Response.record_id == Record.id,
                Response.user_id == user_id,
            ),
        )
        .with_entities(func.count())
        .filter(and_(Record.dataset_id == dataset_id, Response.status == None))
        .scalar()
    )


async def create_response(
    db: Session, search_engine: SearchEngine, record: Record, user: User, response_create: ResponseCreate
):
    validate_response_values(record.dataset, values=response_create.values, status=response_create.status)

    response = Response(
        values=jsonable_encoder(response_create.values),
        status=response_create.status,
        record_id=record.id,
        user_id=user.id,
    )

    db.add(response)
    db.flush([response])
    # TODO: Rollback
    await search_engine.update_record_response(response)

    db.commit()
    db.refresh(response)

    return response


async def update_response(
    db: Session, search_engine: SearchEngine, response: Response, response_update: ResponseUpdate
):
    validate_response_values(response.record.dataset, values=response_update.values, status=response_update.status)

    response.values = jsonable_encoder(response_update.values)
    response.status = response_update.status

    db.flush([response])
    # TODO: Rollback
    await search_engine.update_record_response(response)

    db.commit()
    db.refresh(response)

    return response


async def delete_response(db: Session, search_engine: SearchEngine, response: Response):
    db.delete(response)
    # TODO: Rollback
    await search_engine.delete_record_response(response)

    db.commit()

    return response


def validate_response_values(dataset: Dataset, values: Dict[str, ResponseValue], status: ResponseStatus):
    if not values:
        if status != ResponseStatus.discarded:
            raise ValueError("Missing response values")
        return

    values_copy = copy.copy(values or {})
    for question in dataset.questions:
        if (
            question.required
            and status == ResponseStatus.submitted
            and not (question.name in values and values_copy.get(question.name))
        ):
            raise ValueError(f"Missing required question: {question.name!r}")

        question_response = values_copy.pop(question.name, None)
        if question_response:
            question.parsed_settings.check_response(question_response)

    if values_copy:
        raise ValueError(f"Error: found responses for non configured questions: {list(values_copy.keys())!r}")


def validate_record_fields(dataset: Dataset, fields: Dict[str, Any]):
    fields_copy = copy.copy(fields or {})
    for field in dataset.fields:
        if field.required and not (field.name in fields_copy and fields_copy.get(field.name) is not None):
            raise ValueError(f"Missing required value for field: {field.name!r}")

        value = fields_copy.pop(field.name, None)
        if not isinstance(value, str):
            raise ValueError(
                f"Wrong value found for field {field.name!r}. Expected {str.__name__!r}, found {type(value).__name__!r}"
            )

    if fields_copy:
        raise ValueError(f"Error: found fields values for non configured fields: {list(fields_copy.keys())!r}")


def _count_fields_by_dataset_id(db: Session, dataset_id: UUID):
    return db.query(func.count(Field.id)).filter_by(dataset_id=dataset_id).scalar()


def _count_questions_by_dataset_id(db: Session, dataset_id: UUID):
    return db.query(func.count(Question.id)).filter_by(dataset_id=dataset_id).scalar()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import textwrap
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List, Optional, Union

from argilla.client.feedback.training.schemas import (
    TrainingTaskMappingForTextClassification,
)
from argilla.client.models import Framework, TextClassificationRecord
from argilla.training import ArgillaTrainer as ArgillaTrainerV1

os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

if TYPE_CHECKING:
    import spacy


class ArgillaTrainer(ArgillaTrainerV1):
    def __init__(
        self,
        dataset: "argilla.client.feedback.schemas.FeedbackDataset",
        task_mapping: TrainingTaskMappingForTextClassification,
        framework: Framework,
        lang: Optional["spacy.Language"] = None,
        model: Optional[str] = None,
        train_size: Optional[float] = None,
        seed: Optional[int] = None,
        gpu_id: Optional[int] = -1,
        framework_kwargs: Optional[dict] = {},
        fetch_records: bool = True,
    ) -> None:
        """
        Initialize an Argilla Trainer.

        Args:
            dataset (FeedbackDataset): the dataset to be used for training.
            task_mapping (TrainingTaskMappingForTextClassification): the training data to be used for training.
            framework (str):
                the framework to use for training. Currently, only "transformers", "setfit", and "spacy"
                are supported.
            lang (spacy.Language):
                the spaCy language model to use for training, just required when `framework="spacy"`.
                Defaults to None, but it will be set to `spacy.blank("en")` if not specified.
            model (str):
                name or path to the baseline model to be used. If not specified will set to a good default
                per framework, if applicable. Defaults to None.
            train_size (float):
                the size of the training set. If not specified, the entire dataset will be used for training,
                which may be an issue if `framework="spacy"` as it requires a validation set. Defaults to None.
            seed (int): the random seed to ensure reproducibility. Defaults to None.
            gpu_id (int):
                the GPU ID to use when training a SpaCy model. Defaults to -1, which means that the CPU
                will be used by default. GPU IDs start in 0, which stands for the default GPU in the system,
                if available.
            framework_kwargs (dict): arguments for the framework's trainer.
            **load_kwargs: arguments for the rg.load() function.
        """
        self._dataset = dataset
        self._train_size = train_size
        self._task_mapping = task_mapping
        self._seed = seed  # split is used for train-test-split and should therefore be fixed
        self._model = model

        self._prepared_data = self._dataset.prepare_for_training(
            framework=framework,
            task_mapping=task_mapping,
            fetch_records=fetch_records,
            train_size=train_size,
            seed=seed,
            lang=lang,
        )

        if isinstance(framework, str):
            framework = Framework(framework)

        if framework is Framework.SETFIT:
            if not isinstance(task_mapping, TrainingTaskMappingForTextClassification):
                raise NotImplementedError(f"{Framework.SETFIT} only supports `TextClassification` tasks.")
            from argilla.client.feedback.training.frameworks.setfit import (
                ArgillaSetFitTrainer,
            )

            self._trainer = ArgillaSetFitTrainer(
                feedback_dataset=self._dataset,
                task_mapping=self._task_mapping,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.TRANSFORMERS:
            from argilla.client.feedback.training.frameworks.transformers import (
                ArgillaTransformersTrainer,
            )

            self._trainer = ArgillaTransformersTrainer(
                feedback_dataset=self._dataset,
                task_mapping=self._task_mapping,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.PEFT:
            from argilla.client.feedback.training.frameworks.peft import (
                ArgillaPeftTrainer,
            )

            self._trainer = ArgillaPeftTrainer(
                feedback_dataset=self._dataset,
                task_mapping=self._task_mapping,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.SPACY:
            from argilla.client.feedback.training.frameworks.spacy import (
                ArgillaSpaCyTrainer,
            )

            self._trainer = ArgillaSpaCyTrainer(
                feedback_dataset=self._dataset,
                task_mapping=self._task_mapping,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
                gpu_id=gpu_id,
                framework_kwargs=framework_kwargs,  # freeze_tok2vec
            )
        elif framework is Framework.SPACY_TRANSFORMERS:
            from argilla.client.feedback.training.frameworks.spacy import (
                ArgillaSpaCyTransformersTrainer,
            )

            self._trainer = ArgillaSpaCyTransformersTrainer(
                feedback_dataset=self._dataset,
                task_mapping=self._task_mapping,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
                gpu_id=gpu_id,
                framework_kwargs=framework_kwargs,  # update_transformer
            )
        elif framework is Framework.OPENAI:
            from argilla.client.feedback.training.frameworks.openai import (
                ArgillaOpenAITrainer,
            )

            self._trainer = ArgillaOpenAITrainer(
                feedback_dataset=self._dataset,
                task_mapping=self._task_mapping,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.SPAN_MARKER:
            from argilla.client.feedback.training.frameworks.span_marker import (
                ArgillaSpanMarkerTrainer,
            )

            self._trainer = ArgillaSpanMarkerTrainer(
                feedback_dataset=self._dataset,
                task_mapping=self._task_mapping,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        else:
            raise NotImplementedError(f"{framework} is not a valid framework.")

        self._logger.info(self)

    def __repr__(self) -> str:
        """
        `trainer.__repr__()` prints out the trainer's parameters and a summary of how to use the trainer

        Returns:
          The trainer object.
        """
        return textwrap.dedent(
            f"""\
            ArgillaBaseTrainer info:
            _________________________________________________________________
            These baseline params are fixed:
                dataset: {self._dataset}
                task: {self._task_mapping}
                train_size: {self._train_size}
                seed: {self._seed}


            {self._trainer.__class__} info:
            _________________________________________________________________
            The parameters are configurable via `trainer.update_config()`:
                {self._trainer}

            Using the trainer:
            _________________________________________________________________
            `trainer.train(output_dir)` to train to start training. `output_dir` is the directory to save the model automatically.
            `trainer.predict(text, as_argilla_records=True)` to make predictions.
            `trainer.save(output_dir)` to save the model manually.
            """
        )

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        `predict` takes a string or list of strings and returns a list of dictionaries, each dictionary
        containing the text, the predicted label, and the confidence score.

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the output will be a list of Argilla records instead of dictionaries. Defaults to True.

        Returns:
          A list of predictions or Argilla records.
        """
        return self._trainer.predict(text=text, as_argilla_records=False, **kwargs)


class ArgillaTrainerSkeleton(ABC):
    def __init__(
        self,
        feedback_dataset: "argilla.client.feedback.schemas.FeedbackDataset",
        task_mapping: TrainingTaskMappingForTextClassification,
        prepared_data=None,
        model: str = None,
        seed: int = None,
        *arg,
        **kwargs,
    ):
        self._feedback_dataset = feedback_dataset
        self._task_mapping = task_mapping
        self._dataset = prepared_data
        self._model = model
        self._seed = seed
        if isinstance(self._task_mapping, TrainingTaskMappingForTextClassification):
            self._multi_label = self._task_mapping.__multi_label__ or False
            self._label_list = self._task_mapping.__all_labels__ or None
            self._label2id = self._task_mapping.__label2id__
            self._id2label = self._task_mapping.__id2label__
            self._record_class = TextClassificationRecord  # TODO: dirty hack to inherit from original trainers

    @abstractmethod
    def init_training_args(self):
        """
        Initializes the training arguments.
        """

    @abstractmethod
    def init_model(self):
        """
        Initializes a model.
        """

    @abstractmethod
    def update_config(self, *args, **kwargs):
        """
        Updates the configuration of the trainer, but the parameters depend on the trainer.subclass.
        """

    @abstractmethod
    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        Predicts the label of the text.
        """

    @abstractmethod
    def train(self, output_dir: str = None):
        """
        Trains the model.
        """

    @abstractmethod
    def save(self, output_dir: str):
        """
        Saves the model to the specified path.
        """



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import os
from typing import List, Optional, Union
from uuid import uuid4

from datasets import DatasetDict

from argilla.client.models import TextClassificationRecord, TokenClassificationRecord
from argilla.training.base import ArgillaTrainerSkeleton
from argilla.utils.dependency import require_version


class AutoTrainMixIn:
    def prepare_dataset(self, data_dict: Optional[dict] = {}) -> None:
        """
        This function prepares a dataset for autotrain using a dictionary of data and specific column
        mappings.

        Args:
          data_dict_params: A dictionary that contains additional parameters to be passed
            to the `AutoTrainDataset` constructor. These parameters are optional and can be used to
            customize the dataset preparation process.
        """
        from autotrain.dataset import AutoTrainDataset

        self.dset = AutoTrainDataset(
            task=self.task,
            token=self.HF_TOKEN,
            username=self.AUTOTRAIN_USERNAME,
            project_name=self.project_name,
            **data_dict,
            percent_valid=None,
            column_mapping={
                "text": "text",
                "label": "label",
            },
        )

    def get_project_cost(self):
        """
        This function returns the project cost based on various parameters.

        Returns:
          The `get_project_cost` function is being called with several arguments and its return value is
        being returned by the `get_project_cost` method. The exact return value depends on the
        implementation of the `get_project_cost` function.
        """
        from autotrain.utils import get_project_cost

        return get_project_cost(
            username=self.AUTOTRAIN_USERNAME,
            token=self.HF_TOKEN,
            task=self.task,
            num_samples=self._num_samples,
            num_models=self.trainer_kwargs["autotrain"][0]["num_models"]
            if self._model.lower() == "autotrain"
            else len(self.trainer_kwargs["hub_model"]),
        )

    def initialize_project(self):
        """
        This function initializes a project with a dataset, hub model, and job parameters, and logs
        information about the project and its cost.
        """
        import copy

        from autotrain.project import Project

        self.project = Project(
            dataset=self.dset,
            hub_model=None if self._model.lower() == "autotrain" else self._model,
            job_params=copy.deepcopy(self.get_job_params()),
        )
        self._logger.info(self.project)
        self._logger.info(f"Project cost: {self.get_project_cost()}")

    def get_job_params(self):
        if self._model.lower() == "autotrain":
            job_params = self.trainer_kwargs["autotrain"]
            model_choice = "autotrain"
        else:
            job_params = self.trainer_kwargs["hub_model"]
            model_choice = "HuggingFace Hub"
        if model_choice == "autotrain":
            if len(job_params) > 1:
                raise ValueError("Only one job parameter is allowed for AutoTrain.")
            job_params[0].update({"task": self.task})
        elif model_choice == "HuggingFace Hub":
            for i in range(len(job_params)):
                job_params[i].update({"task": self.task})
        return job_params


class ArgillaAutoTrainTrainer(ArgillaTrainerSkeleton, AutoTrainMixIn):
    _logger = logging.getLogger("ArgillaAutoTrainTrainer")
    _logger.setLevel(logging.INFO)

    try:
        AUTOTRAIN_USERNAME = os.environ["AUTOTRAIN_USERNAME"]
        HF_TOKEN = os.environ["HF_AUTH_TOKEN"]
    except KeyError:
        raise KeyError("Please set the `AUTOTRAIN_USERNAME` and `HF_AUTH_TOKEN` environment variables.")

    require_version("autotrain-advanced")
    require_version("datasets")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.project_name = f"{self._workspace}_{self._name}_{str(uuid4())[:8]}"

        if self._seed:
            self._logger.warning("Setting a seed is not supported by `autotrain-advanced`.")
            self._seed = 42

        if self._model is None:
            self._model = "bert-base-uncased"

        data_dict = {}
        self._num_samples = 0
        if isinstance(self._dataset, DatasetDict):
            self._train_dataset = self._dataset["train"]
            self._eval_dataset = self._dataset["test"]
            data_dict["valid_data"] = [self._eval_dataset.to_pandas()]
            self._num_samples += len(self._eval_dataset)
        else:
            self._train_dataset = self._dataset
            self._eval_dataset = None
            data_dict["valid_data"] = []
        data_dict["train_data"] = [self._train_dataset.to_pandas()]
        self._num_samples += len(self._train_dataset)

        if self._record_class == TextClassificationRecord:
            if self._multi_label:
                raise NotImplementedError(
                    "TextClassificaiton `multi_label=True` is not supported by `autotrain-advanced`."
                )
            elif self._multi_label is False:
                n_classes = len(self._train_dataset.features["label"].names)
                self.task = "text_multi_class_classification" if n_classes > 2 else "text_binary_classification"
                self._id2label = dict(enumerate(self._train_dataset.features["label"].names))
                self._label_list = self._train_dataset.features["label"].names
            self._label2id = {v: k for k, v in self._id2label.items()}
        elif self._record_class == TokenClassificationRecord:
            raise NotImplementedError(
                "`Text2Text` and `TokenClassification` tasks are not supported by `autotrain-advanced`."
            )

        self.init_training_args()

        self.prepare_dataset(data_dict=data_dict)
        self.initialize_project()

    def init_training_args(self):
        from autotrain.params import Params

        if not hasattr(self, "trainer_kwargs"):
            self.trainer_kwargs = {"autotrain": [{}], "hub_model": [{}]}
        for training_type in ["autotrain", "hub_model"]:
            for idx, set_params in enumerate(self.trainer_kwargs[training_type]):
                params = Params(
                    task=self.task,
                    training_type=training_type,
                ).get()
                for key, value in params.items():
                    if key not in set_params:
                        set_params[key] = value.DEFAULT

                self.trainer_kwargs[training_type][idx] = set_params

    def update_config(self, **kwargs):
        """
        Updates the `setfit_model_kwargs` and `setfit_trainer_kwargs` dictionaries with the keyword
        arguments passed to the `update_config` function.
        """
        self._model = kwargs.get("model", self._model)
        if self._model.lower() == "autotrain":
            self._model = "autotrain"
        if "hub_model" in kwargs:
            if not isinstance(kwargs["hub_model"], list):
                raise ValueError("hub_model must be a list of dictionaries.")
            self.trainer_kwargs["hub_model"] = kwargs["hub_model"]
        if "autotrain" in kwargs:
            if not isinstance(kwargs["autotrain"], list):
                raise ValueError("autotrain must be a list of dictionaries.")
            self.trainer_kwargs["autotrain"] = kwargs["autotrain"]

        self.init_training_args()
        self.initialize_project()

    def __repr__(self):
        formatted_string = [
            "Choose EITHER `autotrain` OR a model from the hub `hub_model` as main key to update parameters.",
            f"model: {self._model}",
        ]
        for arg_dict_key, arg_dict_single in self.trainer_kwargs.items():
            arg_dict_key += ": List[dict]"
            formatted_string.append(arg_dict_key)
            for idx, item in enumerate(arg_dict_single):
                for key, val in item.items():
                    formatted_string.append(f"\tjob{idx+1}-{key}: {val}")
        return "\n".join(formatted_string)

    def train(self, output_dir: str):
        """
        We create a SetFitModel object from a pretrained model, then create a SetFitTrainer object with
        the model, and then train the model
        """
        self.project.dataset.prepare()
        project_id = self.project.create()
        self.project.approve(project_id)

    def init_model(self, new: bool = False):
        pass

    def init_pipeline(self):
        pass

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        The function takes in a list of strings and returns a list of predictions

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the prediction will be returned as an Argilla record. If
        False, the prediction will be returned as a string. Defaults to True

        Returns:
          A list of predictions
        """
        self._logger.error(
            f"Use ArgillaTrainer(name={self._name}, workspace={self._workspace}, framework='transformers', model='my_model_name') for inference via `.predict()`."
        )

    def save(self, output_dir: str):
        """
        The function saves the model to the path specified, and also saves the label2id and id2label
        dictionaries to the same path

        Args:
          output_dir (str): the path to save the model to
        """
        self._logger.warning("Models are saved on the HuggingFace Hub, so this function is not supported.")



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

from argilla.server.daos.backend.metrics.base import ElasticsearchMetric
from argilla.server.daos.backend.search.model import BaseQuery, SortConfig


@dataclasses.dataclass
class IClientAdapter(metaclass=ABCMeta):
    vector_search_supported: bool

    @abstractmethod
    def get_cluster_info(self) -> Dict[str, Any]:
        """Returns basic about es cluster"""
        pass

    @abstractmethod
    def get_index_schema(self, *, index: str):
        pass

    @abstractmethod
    def drop_document_property(
        self,
        index: str,
        id: str,
        property: str,
    ):
        pass

    @abstractmethod
    def update_docs_by_query(
        self,
        *,
        index: str,
        data: Dict[str, Any],
        query: Optional[BaseQuery] = None,
    ) -> Dict[str, Any]:
        pass

    @abstractmethod
    def delete_docs_by_query(
        self,
        *,
        index: str,
        query: Optional[BaseQuery],
    ) -> Tuple[int, int]:
        pass

    @abstractmethod
    def scan_docs(
        self,
        index: str,
        query: BaseQuery,
        sort: SortConfig,
        size: Optional[int] = None,
        fetch_once: bool = False,
        search_from_params: Optional[Any] = None,
        enable_highlight: bool = False,
        include_fields: Optional[List[str]] = None,
        exclude_fields: Optional[List[str]] = None,
    ) -> Iterable[Dict[str, Any]]:
        pass

    @abstractmethod
    def search_docs(
        self,
        *,
        index: str,
        query: Optional[BaseQuery] = None,
        sort: Optional[SortConfig] = None,
        doc_from: int = 0,
        size: int = 100,
        exclude_fields: List[str] = None,
        enable_highlight: bool = True,
        routing: str = None,
    ):
        pass

    @abstractmethod
    def copy_index(
        self,
        *,
        source_index: str,
        target_index: str,
        override: bool = True,
        reindex: bool = False,
    ):
        pass

    @abstractmethod
    def get_property_type(
        self,
        *,
        index: str,
        property_name: str,
        drop_extra_props: bool = False,
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def get_index_document_by_id(
        self,
        index: str,
        id: str,
    ) -> Optional[Dict[str, Any]]:
        pass

    @abstractmethod
    def create_index_alias(
        self,
        *,
        index: str,
        alias: str,
    ):
        pass

    @abstractmethod
    def delete_index_document(
        self,
        index: str,
        id: str,
    ):
        pass

    @abstractmethod
    def delete_index_alias(
        self,
        *,
        index: str,
        alias: str,
    ):
        pass

    @abstractmethod
    def exists_index(self, index: str) -> bool:
        pass

    @abstractmethod
    def configure_index_vectors(
        self,
        *,
        index: str,
        vectors: Dict[str, int],
    ):
        pass

    @abstractmethod
    def delete_index(
        self,
        *,
        index: str,
        raises_error: bool = False,
    ):
        pass

    @abstractmethod
    def create_index(
        self,
        index: str,
        force_recreate: bool = False,
        settings: Dict[str, Any] = None,
        mappings: Dict[str, Any] = None,
    ):
        pass

    @abstractmethod
    def index_documents(self, index: str, docs: List[Dict[str, Any]]) -> int:
        pass

    @abstractmethod
    def upsert_index_document(
        self,
        index: str,
        id: str,
        document: Optional[Dict[str, Any]] = None,
        script: Optional[str] = None,
        partial_update: bool = False,
    ):
        pass

    @abstractmethod
    def open_index(self, index: str):
        pass

    @abstractmethod
    def close_index(self, index: str):
        pass

    @abstractmethod
    def clone_index(
        self,
        source_index: str,
        target_index: str,
    ):
        pass

    @abstractmethod
    def set_index_mappings(
        self,
        index: str,
        properties: Optional[Dict[str, Any]] = None,
        **extra_cfg,
    ):
        pass

    @abstractmethod
    def compute_index_metric(
        self,
        *,
        index: str,
        metric: ElasticsearchMetric,
        query: BaseQuery,
        params: Dict[str, Any],
    ):
        pass



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import datetime
import functools
import inspect
import json
import uuid
from json import JSONEncoder
from typing import Dict, Optional, TypeVar
from urllib.parse import urlparse

import httpx

from argilla._constants import API_KEY_HEADER_NAME
from argilla.client.sdk._helpers import build_raw_response
from argilla.client.sdk.commons.errors import BaseClientError


@dataclasses.dataclass
class _ClientCommonDefaults:
    __httpx__: httpx.Client = dataclasses.field(default=None, init=False, compare=False)

    cookies: Dict[str, str] = dataclasses.field(default_factory=dict)
    headers: Dict[str, str] = dataclasses.field(default_factory=dict)
    timeout: float = 5.0

    def get_headers(self) -> Dict[str, str]:
        """Get headers to be used in all endpoints"""
        return {**self.headers}

    def get_cookies(self) -> Dict[str, str]:
        return {**self.cookies}

    def get_timeout(self) -> float:
        return self.timeout

    def update_headers(self, headers: Dict[str, str]):
        self.headers.update(headers)
        self.httpx.headers.update(self.get_headers())

    @property
    def httpx(self):
        return self.__httpx__


@dataclasses.dataclass
class _Client:
    base_url: str

    def __post_init__(self):
        self.base_url = self.base_url.strip()
        if self.base_url.endswith("/"):
            self.base_url = self.base_url[:-1]

        url = urlparse(self.base_url)
        if url.scheme == "https" and "_" in url.hostname:
            self.__httpx__ = None
            raise ValueError(
                'Avoid using hostnames with underscores "_". For reference see:'
                " https://stackoverflow.com/questions/10959757/the-use-of-the-underscore-in-host-names"
            )


@dataclasses.dataclass
class _AuthenticatedClient(_Client):
    token: str


class _EnhancedJSONEncoder(JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if isinstance(o, uuid.UUID):
            return str(o)
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        return super().default(o)


@dataclasses.dataclass
class Client(_ClientCommonDefaults, _Client):
    def __post_init__(self):
        super().__post_init__()
        self.__httpx__ = httpx.Client(
            base_url=self.base_url,
            headers=self.get_headers(),
            cookies=self.get_cookies(),
            timeout=self.get_timeout(),
        )
        # TODO: Remove this NOW!!!!
        self.__http_async__ = httpx.AsyncClient(
            base_url=self.base_url,
            headers=self.get_headers(),
            cookies=self.get_cookies(),
            timeout=self.get_timeout(),
        )

    def __del__(self):
        del self.__httpx__

    def __hash__(self):
        return hash(self.base_url)

    def with_httpx_error_handler(func):
        def wrap_error(base_url: str):
            err_str = f"Your Api endpoint at {base_url} is not available or not responding."
            raise BaseClientError(err_str) from None

        @functools.wraps(func)
        def inner(self, *args, **kwargs):
            try:
                result = func(self, *args, **kwargs)
                return result
            except httpx.ConnectError as err:  # noqa: F841
                return wrap_error(self.base_url)

        return inner

    @with_httpx_error_handler
    def get(self, path: str, *args, **kwargs):
        path = self._normalize_path(path)
        response = self.__httpx__.get(
            url=path,
            headers=self.get_headers(),
            *args,
            **kwargs,
        )
        return build_raw_response(response).parsed

    @with_httpx_error_handler
    def patch(self, path: str, *args, **kwargs):
        path = self._normalize_path(path)

        response = self.__httpx__.patch(
            url=path,
            headers=self.get_headers(),
            *args,
            **kwargs,
        )
        return build_raw_response(response).parsed

    @with_httpx_error_handler
    def patch(
        self,
        path: str,
        *args,
        json: Optional[dict] = None,
        **kwargs,
    ):
        path = self._normalize_path(path)
        body = self._normalize_body(json)
        response = self.__httpx__.patch(
            url=path,
            headers=self.get_headers(),
            json=body,
            *args,
            **kwargs,
        )
        return build_raw_response(response).parsed

    @with_httpx_error_handler
    def post(
        self,
        path: str,
        *args,
        json: Optional[dict] = None,
        **kwargs,
    ):
        path = self._normalize_path(path)
        body = self._normalize_body(json)
        response = self.__httpx__.post(
            url=path,
            headers=self.get_headers(),
            json=body,
            *args,
            **kwargs,
        )
        return build_raw_response(response).parsed

    @with_httpx_error_handler
    def put(
        self,
        path: str,
        *args,
        json: Optional[dict] = None,
        **kwargs,
    ):
        path = self._normalize_path(path)
        body = self._normalize_body(json)
        response = self.__httpx__.put(
            url=path,
            headers=self.get_headers(),
            json=body,
            *args,
            **kwargs,
        )
        return build_raw_response(response).parsed

    @with_httpx_error_handler
    def delete(
        self,
        path: str,
        *args,
        json: Optional[dict] = None,
        **kwargs,
    ):
        path = self._normalize_path(path)
        body = self._normalize_body(json)
        response = self.__httpx__.request(
            method="DELETE",
            url=path,
            headers=self.get_headers(),
            json=body,
            *args,
            **kwargs,
        )
        return build_raw_response(response).parsed

    @staticmethod
    def _normalize_path(path: str) -> str:
        path = path.strip()
        if path.startswith("/"):
            return path[1:]
        return path

    @classmethod
    def _normalize_body(cls, body: dict) -> dict:
        json_str = json.dumps(
            body,
            cls=_EnhancedJSONEncoder,
        )
        return json.loads(json_str)


ResponseType = TypeVar("ResponseType")


@dataclasses.dataclass
class AuthenticatedClient(
    Client,
    _ClientCommonDefaults,
    _AuthenticatedClient,
):
    """A Client which has been authenticated for use on secured endpoints"""

    def __hash__(self):
        return hash((self.base_url, self.token))

    def get_headers(self) -> Dict[str, str]:
        """Get headers to be used in authenticated endpoints"""
        return {
            "Authorization": f"Bearer {self.token}",  # Backward compatibility
            API_KEY_HEADER_NAME: self.token,
            **super().get_headers(),
        }



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import TYPE_CHECKING, Iterator, List, Optional, Union
from uuid import UUID

from argilla.client import active_client
from argilla.client.sdk.commons.errors import (
    AlreadyExistsApiError,
    BaseClientError,
    NotFoundApiError,
    ValidationApiError,
)
from argilla.client.sdk.v1.workspaces import api as workspaces_api_v1
from argilla.client.sdk.v1.workspaces.models import WorkspaceModel as WorkspaceModelV1
from argilla.client.sdk.workspaces import api as workspaces_api
from argilla.client.sdk.workspaces.models import WorkspaceModel as WorkspaceModelV0

if TYPE_CHECKING:
    import httpx

    from argilla.client.sdk.workspaces.models import WorkspaceUserModel


class Workspace:
    """The `Workspace` class is used to manage workspaces in Argilla. It provides
    methods to create new workspaces, adding users to them, listing the linked users,
    and deleting users from the workspace. While it's not allowed to delete a workspace
    neither to update the workspace name.

    Args:
        name: the name of the workspace to be managed. Defaults to None.
        id: the ID of the workspace to be managed. Defaults to None.

    Attributes:
        __client: the `httpx.Client` initialized to interact with the Argilla API.
        id: the ID of the workspace.
        name: the name of the workspace.
        users: the list of users linked to the workspace. Defaults to None.
        inserted_at: the datetime when the workspace was created.
        updated_at: the datetime when the workspace was last updated.

    Examples:
        >>> from argilla import rg
        >>> workspace = rg.Workspace.from_name("my-workspace") # or `Workspace.from_id("...")`
        >>> workspace.add_user("my-user")
        >>> print(workspace.users)
        [WorkspaceUserModel(id='...', first_name='Luke', last_name="Skywalker', full_name='Luke Skywalker', username='my-user', role='annotator', workspaces=['my-workspace'], api_key='...', inserted_at=datetime.datetime(2021, 8, 31, 10, 0, 0), updated_at=datetime.datetime(2021, 8, 31, 10, 0, 0))]
        >>> workspace.delete_user("my-user")
        >>> print(workspace.users)
        []
    """

    __client: "httpx.Client"
    id: UUID
    name: str
    users: Optional[List["WorkspaceUserModel"]] = None
    inserted_at: datetime
    updated_at: datetime

    def __init__(
        self,
        name: Optional[str] = None,
        *,
        id: Optional[Union[str, UUID]] = None,
    ) -> None:
        """Doesn't initialize the `Workspace`, as it needs to be instantiated via any of
        the following classmethods, as the Argilla API is required.
            - `Workspace.create` to create a new workspace in Argilla
            - `Workspace.from_name` to get an existing workspace from Argilla by its name
            - `Workspace.from_id` to get an existing workspace from Argilla by its ID
            - `Workspace.list` to list all the workspaces in Argilla

        Args:
            name: the name of the workspace to be managed. Defaults to None.
            id: the ID of the workspace to be managed. Defaults to None.

        Raises:
            Exception: if the `Workspace` is initialized via the `__init__` method.
        """
        error_msg = (
            "`Workspace` cannot be initialized via the `__init__` method, as it needs"
            " to send requests to the Argilla API, so it can just be instantiated via"
            " the following classmethods: `Workspace.create`, `Workspace.from_name`,"
            " `Workspace.from_id` and `Workspace.list`; note that it must exist in"
            " advance in Argilla."
        )
        if id is not None:
            error_msg += f" As the `id` argument is not None, you should use `Workspace.from_id('{id}')` instead."
        if name is not None:
            error_msg += f" As the `name` argument is not None, you should use `Workspace.from_name('{name}')` instead."
        raise Exception(error_msg)

    @property
    def users(self) -> List["WorkspaceUserModel"]:
        """Returns the list of users linked to the workspace.

        Returns:
            A list of `WorkspaceUserModel` instances.
        """
        # TODO(@alvarobartt): Maybe we should return a list of rg.User instead.
        return workspaces_api.list_workspace_users(self.__client, self.id).parsed

    def __repr__(self) -> str:
        return (
            f"Workspace(id={self.id}, name={self.name},"
            f" inserted_at={self.inserted_at}, updated_at={self.updated_at})"
        )

    def add_user(self, user_id: str) -> None:
        """Adds an existing user to the workspace in Argilla.

        Args:
            user_id: the ID of the user to be added to the workspace. The user must exist in Argilla.

        Raises:
            ValueError: if the user with the provided ID already exists in the workspace.
            RuntimeError: if there was an error while adding the user to the workspace.

        Examples:
            >>> from argilla import rg
            >>> workspace = rg.Workspace.from_name("my-workspace")
            >>> workspace.add_user("my-user-id")
        """
        try:
            workspaces_api.create_workspace_user(
                client=self.__client,
                id=self.id,
                user_id=user_id,
            )
        except AlreadyExistsApiError as e:
            raise ValueError(f"User with id=`{user_id}` already exists in workspace with id=`{self.id}`.") from e
        except BaseClientError as e:
            raise RuntimeError(f"Error while adding user with id=`{user_id}` to workspace with id=`{self.id}`.") from e

    def delete_user(self, user_id: str) -> None:
        """Deletes an existing user from the workspace in Argilla. Note that the user
        will not be deleted from Argilla, but just from the workspace.

        Args:
            user_id: the ID of the user to be deleted from the workspace. The user must exist in Argilla.

        Raises:
            ValueError: if the user with the provided ID doesn't exist in the workspace.
            RuntimeError: if there was an error while deleting the user from the workspace.

        Examples:
            >>> from argilla import rg
            >>> workspace = rg.Workspace.from_name("my-workspace")
            >>> workspace.delete_user("my-user-id")
        """
        try:
            workspaces_api.delete_workspace_user(
                client=self.__client,
                id=self.id,
                user_id=user_id,
            )
        except NotFoundApiError as e:
            raise ValueError(
                f"Either the user with id=`{user_id}` doesn't exist in Argilla, or it"
                f" doesn't belong to workspace with id=`{self.id}`."
            ) from e
        except BaseClientError as e:
            raise RuntimeError(
                f"Error while deleting user with id=`{user_id}` from workspace with id=`{self.id}`."
            ) from e

    @staticmethod
    def __active_client() -> "httpx.Client":
        """Returns the active Argilla `httpx.Client` instance."""
        try:
            return active_client().http_client.httpx
        except Exception as e:
            raise RuntimeError(f"The `rg.active_client()` is not available or not respoding.") from e

    @classmethod
    def __new_instance(
        cls, client: Optional["httpx.Client"] = None, ws: Optional[Union[WorkspaceModelV0, WorkspaceModelV1]] = None
    ) -> "Workspace":
        """Returns a new `Workspace` instance."""
        instance = cls.__new__(cls)
        instance.__client = client or cls.__active_client()
        if isinstance(ws, (WorkspaceModelV0, WorkspaceModelV1)):
            instance.__dict__.update(ws.dict())
        return instance

    @classmethod
    def create(cls, name: str) -> "Workspace":
        """Creates a new workspace in Argilla.

        Args:
            name: the name of the workspace to be created.

        Raises:
            ValueError: if the workspace with the provided name already exists.

        Returns:
            A `Workspace` instance.

        Examples:
            >>> from argilla import rg
            >>> workspace = rg.Workspace.create("my-workspace")
        """
        client = cls.__active_client()
        try:
            ws = workspaces_api.create_workspace(client, name).parsed
            return cls.__new_instance(client, ws)
        except AlreadyExistsApiError as e:
            raise ValueError(f"Workspace with name=`{name}` already exists, so please use a different name.") from e
        except (ValidationApiError, BaseClientError) as e:
            raise RuntimeError(f"Error while creating workspace with name=`{name}`.") from e

    @classmethod
    def from_id(cls, id: UUID) -> "Workspace":
        """Gets an existing workspace from Argilla by its ID.

        Args:
            id: the ID of the workspace to be retrieved.

        Returns:
            A `Workspace` instance.

        Raises:
            ValueError: if the workspace with the provided ID doesn't exist, or if it's
                not a valid UUID identifier.
            RuntimeError: if there was an error while retrieving the workspace.

        Examples:
            >>> from argilla import rg
            >>> workspace = rg.Workspace.from_id("my-workspace-id")
        """
        client = cls.__active_client()
        try:
            ws = workspaces_api_v1.get_workspace(client, id).parsed
            return cls.__new_instance(client, ws)
        except NotFoundApiError as e:
            raise ValueError(
                f"Workspace with id=`{id}` doesn't exist in Argilla, so please"
                " make sure that the ID you provided is a valid one. Otherwise,"
                " you can create a new one via the `Workspace.create` method."
            ) from e
        except ValidationApiError as e:
            raise ValueError(
                "The ID you provided is not a valid UUID, so please make sure that the"
                " ID you provided is a valid one."
            ) from e
        except BaseClientError as e:
            raise RuntimeError(f"Error while retrieving workspace with id=`{id}` from Argilla.") from e

    @classmethod
    def from_name(cls, name: str) -> "Workspace":
        """Gets an existing workspace from Argilla by its name.

        Args:
            name: the name of the workspace to be retrieved.

        Returns:
            A `Workspace` instance.

        Raises:
            RuntimeError: if there was an error while listing the workspaces.
            ValueError: if the workspace with the provided name doesn't exist.

        Examples:
            >>> from argilla import rg
            >>> workspace = rg.Workspace.from_name("my-workspace")
        """
        client = cls.__active_client()
        try:
            workspaces = workspaces_api.list_workspaces(client).parsed
        except Exception as e:
            raise RuntimeError("Error while retrieving the list of workspaces from Argilla.") from e

        for ws in workspaces:
            if ws.name == name:
                return cls.__new_instance(client, ws)

        raise ValueError(
            f"Workspace with name=`{name}` doesn't exist in Argilla, so please"
            "create it via the `Workspace.create` method as follows:"
            f" `Workspace.create('{name}')`."
        )

    @classmethod
    def list(cls) -> Iterator["Workspace"]:
        """Lists all the workspaces in Argilla.

        Returns:
            A list of `Workspace` instances.

        Raises:
            RuntimeError: if there was an error while listing the workspaces.

        Examples:
            >>> from argilla import rg
            >>> workspaces = rg.Workspace.list()
        """
        client = cls.__active_client()
        try:
            workspaces = workspaces_api.list_workspaces(client).parsed
            for ws in workspaces:
                yield cls.__new_instance(client, ws)
        except Exception as e:
            raise RuntimeError("Error while retrieving the list of workspaces from Argilla.") from e



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy.orm import Session

from argilla.server.contexts import accounts
from argilla.server.database import get_db
from argilla.server.policies import WorkspacePolicyV1, authorize
from argilla.server.schemas.v1.workspaces import Workspace
from argilla.server.security import auth
from argilla.server.security.model import User

router = APIRouter(tags=["workspaces"])


@router.get("/workspaces/{workspace_id}", response_model=Workspace)
def get_workspace(
    *,
    db: Session = Depends(get_db),
    workspace_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    authorize(current_user, WorkspacePolicyV1.get(workspace_id))

    workspace = accounts.get_workspace_by_id(db, workspace_id)
    if not workspace:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Workspace with id `{workspace_id}` not found",
        )

    return workspace



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io

from alembic import command
from alembic.config import Config


def get_current_revision(alembic_config_file: str, verbose: bool = False) -> str:
    output_buffer = io.StringIO()
    alembic_cfg = Config(alembic_config_file, stdout=output_buffer)

    command.current(alembic_cfg, verbose=verbose)
    command_result = output_buffer.getvalue().strip()

    if verbose:
        return command_result
    return command_result.split(" ")[0]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional

from pydantic import BaseModel

from argilla.client.feedback.typing import AllowedFieldTypes, AllowedQuestionTypes


class FeedbackDatasetConfig(BaseModel):
    """`FeedbackDatasetConfig`

    Args:
        fields (List[AllowedFieldTypes]): The fields of the feedback dataset.
        questions (List[AllowedQuestionTypes]): The questions of the feedback dataset.
        guidelines (Optional[str]): the guidelines of the feedback dataset. Defaults to None.

    Examples:
        >>> import argilla as rg
        >>> config = rg.FeedbackDatasetConfig(
        ...     fields=[
        ...         rg.TextField(name="text", title="Human prompt"),
        ...     ],
        ...     questions =[
        ...         rg.TextQuestion(
        ...             name="question-1",
        ...             description="This is the first question",
        ...             required=True,
        ...         ),
        ...         rg.RatingQuestion(
        ...             name="question-2",
        ...             description="This is the second question",
        ...             required=True,
        ...             values=[1, 2, 3, 4, 5],
        ...         ),
        ...         rg.LabelQuestion(
        ...             name="relevant",
        ...             title="Is the response relevant for the given prompt?",
        ...             labels=["Yes","No"],
        ...             required=True,
        ...             visible_labels=None
        ...         ),
        ...         rg.MultiLabelQuestion(
        ...             name="content_class",
        ...             title="Does the response include any of the following?",
        ...             description="Select all that apply",
        ...             labels={"cat-1": "Category 1" , "cat-2": "Category 2"},
        ...             required=False,
        ...             visible_labels=4
        ...         ),
        ...     ],
        ...     guidelines="Add some guidelines for the annotation team here."
        ... )

    """

    fields: List[AllowedFieldTypes]
    questions: List[AllowedQuestionTypes]
    guidelines: Optional[str] = None

    class Config:
        smart_union = True



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, Iterable, List, Optional, Tuple

from pydantic import BaseModel, Field

from argilla.logging import LoggingMixin
from argilla.server.commons.models import TaskType
from argilla.server.daos.backend.base import IndexNotFoundError, InvalidSearchError
from argilla.server.daos.backend.client_adapters.base import IClientAdapter
from argilla.server.daos.backend.client_adapters.factory import ClientAdapterFactory
from argilla.server.daos.backend.mappings.datasets import (
    DATASETS_INDEX_NAME,
    datasets_index_mappings,
)
from argilla.server.daos.backend.mappings.helpers import (
    mappings,
    tasks_common_mappings,
    tasks_common_settings,
)
from argilla.server.daos.backend.mappings.text2text import text2text_mappings
from argilla.server.daos.backend.mappings.text_classification import (
    text_classification_mappings,
)
from argilla.server.daos.backend.mappings.token_classification import (
    token_classification_mappings,
)
from argilla.server.daos.backend.metrics import ALL_METRICS
from argilla.server.daos.backend.metrics.base import ElasticsearchMetric
from argilla.server.daos.backend.search.model import (
    BackendRecordsQuery,
    BaseDatasetsQuery,
    BaseQuery,
    BaseRecordsQuery,
    SortableField,
    SortConfig,
)
from argilla.server.errors import BadRequestError, EntityNotFoundError
from argilla.server.errors.task_errors import MetadataLimitExceededError
from argilla.server.settings import settings

NON_SEARCHABLE_PREFIX = "_"


def dataset_records_index(dataset_id: str) -> str:
    index_mame_template = settings.dataset_records_index_name
    return index_mame_template.format(dataset_id)


class PaginatedSortInfo(BaseModel):
    shuffle: bool = False
    sort_by: List[SortableField] = Field(default_factory=list)
    next_search_params: Optional[Any] = None


class GenericElasticEngineBackend(LoggingMixin):
    """
    Encapsulates logic about the communication, queries and index mapping
    transformations between DAOs layer and the elasticsearch backend.
    """

    _INSTANCE = None

    # TODO(@frascuchon): Once id is included as keyword in datasets index, we can discard this
    __MAX_NUMBER_OF_LISTED_DATASETS__ = 2500

    @classmethod
    def get_instance(cls) -> "GenericElasticEngineBackend":
        """
        Creates an instance of ElasticsearchBackend.

        This function is used in fastapi for resolve component dependencies.

        See <https://fastapi.tiangolo.com/tutorial/dependencies/>

        Returns
        -------

        """

        if not cls._INSTANCE:
            instance = cls(
                client=ClientAdapterFactory.get(
                    hosts=settings.elasticsearch,
                    index_shards=settings.es_records_index_shards,
                    ssl_verify=settings.elasticsearch_ssl_verify,
                    ca_path=settings.elasticsearch_ca_path,
                ),
                metrics={**ALL_METRICS},
                mappings={
                    TaskType.text_classification: text_classification_mappings(),
                    TaskType.token_classification: token_classification_mappings(),
                    TaskType.text2text: text2text_mappings(),
                },
            )
            cls._INSTANCE = instance

        return cls._INSTANCE

    def __init__(
        self,
        client: IClientAdapter,
        metrics: Dict[str, ElasticsearchMetric] = None,
        mappings: Dict[str, Dict[str, Any]] = None,
    ):
        self.__client__ = client
        self.__defined_metrics__ = metrics or {}
        self.__tasks_mappings__ = mappings

        self._common_records_mappings = tasks_common_mappings()
        self._common_records_settings = tasks_common_settings()

    @property
    def client(self) -> IClientAdapter:
        """The elasticsearch client"""
        return self.__client__

    def find_metric_by_id(self, metric_id: str) -> Optional[ElasticsearchMetric]:
        metric = self.__defined_metrics__.get(metric_id)
        if not metric:
            raise EntityNotFoundError(
                name=metric_id,
                type=ElasticsearchMetric,
            )
        return metric

    def get_task_mapping(self, task: TaskType) -> Dict[str, Any]:
        return self.__tasks_mappings__[task]

    def compute_metric(
        self,
        id: str,
        metric_id: str,
        query: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        index = dataset_records_index(id)
        return self._compute_metric(
            index=index,
            metric_id=metric_id,
            query=query,
            params=params,
        )

    def get_schema(self, id: str) -> Dict[str, Any]:
        index = dataset_records_index(id)
        return self.client.get_index_schema(index=index)

    async def update_records_content(
        self,
        id: str,
        content: Dict[str, Any],
        query: Optional[BaseDatasetsQuery],
    ) -> Tuple[int, int]:
        index = dataset_records_index(id)
        response = self.client.update_docs_by_query(
            index=index,
            data=content,
            query=query,
        )
        total, updated = response["total"], response["updated"]
        return total, updated

    async def delete_records_by_query(
        self,
        id: str,
        query: Optional[BaseDatasetsQuery],
    ) -> Tuple[int, int]:
        index = dataset_records_index(id)
        total, deleted = self.client.delete_docs_by_query(
            index=index,
            query=query,
        )
        return total, deleted

    def search_records(
        self,
        id: str,
        query: BackendRecordsQuery = BaseRecordsQuery(),
        sort: SortConfig = SortConfig(),
        record_from: int = 0,
        size: int = 100,
        exclude_fields: List[str] = None,
        enable_highlight: bool = True,
    ) -> Tuple[int, List[Dict[str, Any]]]:
        index = dataset_records_index(id)

        if not sort.sort_by and sort.shuffle is False:
            sort.sort_by = [SortableField(id="id")]  # Default sort by id

        results = self.client.search_docs(
            index=index,
            query=query,
            sort=sort,
            doc_from=record_from,
            size=size,
            exclude_fields=exclude_fields,
            enable_highlight=enable_highlight,
        )

        total = results["total"]
        records = results["docs"]

        return total, records

    def scan_records(
        self,
        id: str,
        query: BackendRecordsQuery,
        sort: PaginatedSortInfo,
        limit: Optional[int] = None,
        include_fields: Optional[List[str]] = None,
        exclude_fields: Optional[List[str]] = None,
    ) -> Iterable[Dict[str, Any]]:
        index = dataset_records_index(id)

        yield from self.client.scan_docs(
            index=index,
            query=query,
            sort=SortConfig(shuffle=sort.shuffle, sort_by=sort.sort_by),
            size=limit,
            search_from_params=sort.next_search_params,
            fetch_once=sort.shuffle,
            include_fields=include_fields,
            exclude_fields=exclude_fields,
            enable_highlight=True,
        )

    def open(self, id: str):
        self.client.open_index(dataset_records_index(id))

    def create_dataset(
        self,
        id: str,
        task: TaskType,
        metadata_values: Optional[Dict[str, Any]] = None,
        vectors_cfg: Optional[Dict[str, Any]] = None,
        force_recreate: bool = False,
    ) -> None:
        _mappings = self._common_records_mappings
        task_mappings = self.get_task_mapping(task).copy()
        for k in task_mappings:
            if isinstance(task_mappings[k], list):
                _mappings[k] = [*_mappings.get(k, []), *task_mappings[k]]
            else:
                _mappings[k] = {**_mappings.get(k, {}), **task_mappings[k]}

        index = dataset_records_index(id)
        self.client.create_index(
            index=index,
            settings=self._common_records_settings,
            mappings={**self._common_records_mappings, **_mappings},
            force_recreate=force_recreate,
        )
        if metadata_values:
            self._configure_metadata_fields(
                index=index,
                metadata_values=metadata_values,
            )

        if vectors_cfg:
            self._configure_vectors_fields(
                index=index,
                vectors_cfg=vectors_cfg,
            )

    def _configure_vectors_fields(
        self,
        index: str,
        vectors_cfg: dict,
    ):
        def _check_max_number_of_vectors():
            vectors = self.client.get_property_type(
                index=index,
                property_name="vectors",
                drop_extra_props=True,
            )
            vector_names = {key for key in vectors.keys() if "vector" in vectors[key]}
            for key in vectors_cfg.keys():
                vector_names.add(f"vectors.{key}.value")

            if len(vector_names) > settings.vectors_fields_limit:
                raise BadRequestError(
                    detail=f"Cannot create more than {settings.vectors_fields_limit} " "kind of vectors per dataset"
                )

        _check_max_number_of_vectors()

        self.client.configure_index_vectors(
            index=index,
            vectors=vectors_cfg,
        )

    def _configure_metadata_fields(self, index: str, metadata_values: Dict[str, Any]):
        def check_metadata_length(metadata_length: int = 0):
            if metadata_length > settings.metadata_fields_limit:
                raise MetadataLimitExceededError(
                    length=metadata_length,
                    limit=settings.metadata_fields_limit,
                )

        def detect_nested_type(v: Any) -> bool:
            """Returns True if value match as nested value"""
            return isinstance(v, list) and isinstance(v[0], dict)

        check_metadata_length(len(metadata_values))
        check_metadata_length(
            len(
                {
                    *self.client.get_property_type(
                        index=index,
                        property_name="metadata",
                        drop_extra_props=True,
                    ),
                    *[f"metadata.{k}" for k in metadata_values.keys()],
                }
            )
        )

        index_mappings = {}
        for field, value in metadata_values.items():
            if field.startswith(NON_SEARCHABLE_PREFIX):
                index_mappings[f"metadata.{field}"] = mappings.non_searchable_text_field()
            elif detect_nested_type(value):
                index_mappings[f"metadata.{field}"] = mappings.nested_field()

        self.client.set_index_mappings(index, properties=index_mappings)

    def delete(self, id: str):
        index = dataset_records_index(id)
        try:
            self.client.delete_index(
                index=index,
                raises_error=True,
            )
        except InvalidSearchError:
            # It's an alias --> DELETE from original index
            original_index = self._old_dataset_index(id)
            self.client.delete_index_alias(
                index=original_index,
                alias=index,
            )
        finally:
            # TODO: This should be move to the service layer and drop from datasets
            self.client.delete_index_document(
                index=DATASETS_INDEX_NAME,
                id=id,
            )

    def _old_dataset_index(self, id):
        return settings.old_dataset_records_index_name.format(id)

    def copy(self, id_from: str, id_to: str):
        index_from = dataset_records_index(id_from)
        index_to = dataset_records_index(id_to)

        self.client.copy_index(
            source_index=index_from,
            target_index=index_to,
        )

    def close(self, id: str):
        return self.client.close_index(index=dataset_records_index(id))

    def create_datasets_index(self, force_recreate: bool = False):
        self.client.create_index(
            index=DATASETS_INDEX_NAME,
            force_recreate=force_recreate,
            mappings=datasets_index_mappings(),
        )
        # TODO: Remove this section of code
        if settings.enable_migration:
            try:
                self._migrate_from_rubrix()
            except IndexNotFoundError:
                pass  # Nothing to migrate

    def _migrate_from_rubrix(self):
        source_index = settings.old_dataset_index_name
        target_index = DATASETS_INDEX_NAME

        try:
            self.client.copy_index(
                source_index=source_index,
                target_index=target_index,
                reindex=True,
            )
            for doc in self.client.scan_docs(index=source_index, query=BaseQuery(), sort=SortConfig()):
                dataset_id = doc["id"]
                index = self._old_dataset_index(dataset_id)
                alias = dataset_records_index(dataset_id=dataset_id)
                self._update_dynamic_mapping(index)
                self.client.create_index_alias(
                    index=index,
                    alias=alias,
                )
        except IndexNotFoundError:
            pass  # Nothing to migrate

    def _update_dynamic_mapping(self, index: str):
        self.client.set_index_mappings(
            index=index,
            dynamic=False,
        )

    def list_datasets(self, query: BaseDatasetsQuery):
        return self.client.scan_docs(
            index=DATASETS_INDEX_NAME,
            query=query,
            sort=SortConfig(),
            fetch_once=True,
            size=self.__MAX_NUMBER_OF_LISTED_DATASETS__,
        )

    def find_record_by_id(self, dataset_id: str, record_id: str):
        index = dataset_records_index(dataset_id)
        return self.client.get_index_document_by_id(
            index=index,
            id=record_id,
        )

    def add_dataset_document(
        self,
        id: str,
        document: Dict[str, Any],
    ):
        self.client.upsert_index_document(
            index=DATASETS_INDEX_NAME,
            id=id,
            document=document,
        )

    def update_dataset_document(
        self,
        id: str,
        document: Dict[str, Any],
    ):
        self.client.upsert_index_document(
            index=DATASETS_INDEX_NAME,
            id=id,
            document=document,
            partial_update=True,
        )

    def update_record(
        self,
        dataset_id: str,
        record_id: str,
        content: dict,
    ):
        index = dataset_records_index(dataset_id)
        self.client.upsert_index_document(
            index=index,
            id=record_id,
            document=content,
        )

    def find_dataset(
        self,
        id: str,
    ):
        document = self.client.get_index_document_by_id(
            index=DATASETS_INDEX_NAME,
            id=id,
        )
        return document

    def _compute_metric(
        self,
        index: str,
        metric_id: str,
        query: Optional[BaseQuery] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict:
        metric = self.find_metric_by_id(metric_id)
        data = self.client.compute_index_metric(
            index=index,
            metric=metric,
            query=query,
            params=params or {},
        )
        return data

    def remove_dataset_field(self, id: str, field: str):
        self.client.drop_document_property(
            index=DATASETS_INDEX_NAME,
            id=id,
            property=field,
        )

    def add_dataset_records(self, id: str, documents: List[dict]) -> int:
        index = dataset_records_index(id)

        return self.client.index_documents(index=index, docs=documents)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

from passlib.context import CryptContext
from sqlalchemy.orm import Session

from argilla.server.models import User, Workspace, WorkspaceUser
from argilla.server.security.model import (
    UserCreate,
    WorkspaceCreate,
    WorkspaceUserCreate,
)

_CRYPT_CONTEXT = CryptContext(schemes=["bcrypt"], deprecated="auto")


def get_workspace_user_by_workspace_id_and_user_id(db: Session, workspace_id: UUID, user_id: UUID):
    return db.query(WorkspaceUser).filter_by(workspace_id=workspace_id, user_id=user_id).first()


def create_workspace_user(db: Session, workspace_user_create: WorkspaceUserCreate):
    workspace_user = WorkspaceUser(
        workspace_id=workspace_user_create.workspace_id,
        user_id=workspace_user_create.user_id,
    )

    db.add(workspace_user)
    db.commit()
    db.refresh(workspace_user)

    return workspace_user


def delete_workspace_user(db: Session, workspace_user: WorkspaceUser):
    db.delete(workspace_user)
    db.commit()

    return workspace_user


def get_workspace_by_id(db: Session, workspace_id: UUID):
    return db.get(Workspace, workspace_id)


def get_workspace_by_name(db: Session, workspace_name: str):
    return db.query(Workspace).filter_by(name=workspace_name).first()


def list_workspaces(db: Session):
    return db.query(Workspace).order_by(Workspace.inserted_at.asc()).all()


def create_workspace(db: Session, workspace_create: WorkspaceCreate):
    workspace = Workspace(name=workspace_create.name)

    db.add(workspace)
    db.commit()
    db.refresh(workspace)

    return workspace


def delete_workspace(db: Session, workspace: Workspace):
    db.delete(workspace)
    db.commit()

    return workspace


def get_user_by_id(db: Session, user_id: UUID):
    return db.get(User, user_id)


def get_user_by_username(db: Session, username: str):
    return db.query(User).filter_by(username=username).first()


def get_user_by_api_key(db: Session, api_key: str):
    return db.query(User).filter_by(api_key=api_key).first()


def list_users(db: Session):
    return db.query(User).order_by(User.inserted_at.asc()).all()


def create_user(db: Session, user_create: UserCreate):
    user = User(
        first_name=user_create.first_name,
        last_name=user_create.last_name,
        username=user_create.username,
        role=user_create.role,
        password_hash=hash_password(user_create.password),
    )

    db.add(user)
    db.commit()
    db.refresh(user)

    return user


def delete_user(db: Session, user: User):
    db.delete(user)
    db.commit()

    return user


def authenticate_user(db: Session, username: str, password: str):
    user = get_user_by_username(db, username)

    if user and verify_password(password, user.password_hash):
        return user
    elif user:
        return
    else:
        _CRYPT_CONTEXT.dummy_verify()


def hash_password(password: str):
    return _CRYPT_CONTEXT.hash(password)


def verify_password(password: str, password_hash: str):
    return _CRYPT_CONTEXT.verify(password, password_hash)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from typing import Any, AsyncGenerator, Dict, Iterable, List, Optional, Union
from uuid import UUID

from opensearchpy import AsyncOpenSearch, helpers
from pydantic import BaseModel
from pydantic.utils import GetterDict

from argilla.server.enums import ResponseStatusFilter
from argilla.server.models import (
    Dataset,
    Field,
    FieldType,
    Question,
    QuestionType,
    Record,
    Response,
    ResponseStatus,
    User,
)
from argilla.server.settings import settings


class SearchDocumentGetter(GetterDict):
    def get(self, key: Any, default: Any = None) -> Any:
        if key == "responses":
            return {
                response.user.username: UserResponse(
                    values={k: v["value"] for k, v in response.values.items()} if response.values else None,
                    status=response.status,
                )
                for response in self._obj.responses
            }
        return super().get(key, default)


class UserResponse(BaseModel):
    values: Optional[Dict[str, Any]]
    status: ResponseStatus


class SearchDocument(BaseModel):
    id: UUID
    fields: Dict[str, Any]

    responses: Optional[Dict[str, UserResponse]]

    class Config:
        orm_mode = True
        getter_dict = SearchDocumentGetter


@dataclasses.dataclass
class TextQuery:
    q: str
    field: Optional[str] = None


@dataclasses.dataclass
class Query:
    text: TextQuery


@dataclasses.dataclass
class UserResponseStatusFilter:
    user: User
    status: ResponseStatusFilter


@dataclasses.dataclass
class SearchResponseItem:
    record_id: UUID
    score: Optional[float]


@dataclasses.dataclass
class SearchResponses:
    items: List[SearchResponseItem]
    total: int = 0


@dataclasses.dataclass
class SearchEngine:
    config: Dict[str, Any]

    es_number_of_shards: int
    es_number_of_replicas: int

    def __post_init__(self):
        self.client = AsyncOpenSearch(**self.config)

    async def create_index(self, dataset: Dataset):
        mappings = {
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/dynamic.html#dynamic-parameters
            "dynamic": "strict",
            "dynamic_templates": self._dynamic_templates_for_question_responses(dataset.questions),
            "properties": {
                # See https://www.elastic.co/guide/en/elasticsearch/reference/current/explicit-mapping.html
                "id": {"type": "keyword"},
                "responses": {"dynamic": True, "type": "object"},
                **self._mapping_for_fields(dataset.fields),
            },
        }

        settings = {
            "number_of_shards": self.es_number_of_shards,
            "number_of_replicas": self.es_number_of_replicas,
        }

        index_name = self._index_name_for_dataset(dataset)
        await self.client.indices.create(index=index_name, body=dict(settings=settings, mappings=mappings))

    async def delete_index(self, dataset: Dataset):
        index_name = self._index_name_for_dataset(dataset)
        await self.client.indices.delete(index_name, ignore=[404], ignore_unavailable=True)

    async def add_records(self, dataset: Dataset, records: Iterable[Record]):
        index_name = await self._get_index_or_raise(dataset)

        bulk_actions = [
            {
                "_op_type": "index",  # If document exist, we update source with latest version
                "_id": record.id,
                "_index": index_name,
                **SearchDocument.from_orm(record).dict(),
            }
            for record in records
        ]

        _, errors = await helpers.async_bulk(client=self.client, actions=bulk_actions, raise_on_error=False)
        if errors:
            raise RuntimeError(errors)

    async def update_record_response(self, response: Response):
        record = response.record
        index_name = await self._get_index_or_raise(record.dataset)

        es_response = UserResponse(
            values={k: v["value"] for k, v in response.values.items()} if response.values else None,
            status=response.status,
        )

        await self.client.update(
            index=index_name,
            id=record.id,
            body={"doc": {"responses": {response.user.username: es_response.dict()}}},
        )

    async def delete_record_response(self, response: Response):
        record = response.record
        index_name = await self._get_index_or_raise(record.dataset)

        await self.client.update(
            index=index_name,
            id=record.id,
            body={"script": f'ctx._source["responses"].remove("{response.user.username}")'},
        )

    async def search(
        self,
        dataset: Dataset,
        query: Union[Query, str],
        user_response_status_filter: Optional[UserResponseStatusFilter] = None,
        offset: int = 0,
        limit: int = 100,
    ) -> SearchResponses:
        # See https://www.elastic.co/guide/en/elasticsearch/reference/current/search-search.html

        if isinstance(query, str):
            query = Query(text=TextQuery(q=query))

        text_query = self._text_query_builder(dataset, text=query.text)

        bool_query: dict = {"must": [text_query]}
        if user_response_status_filter:
            bool_query["filter"] = self._response_status_filter_builder(user_response_status_filter)

        body = {
            "_source": False,
            "query": {"bool": bool_query},
            # "sort": [{"_score": "desc"}, {"id": "asc"}],
        }

        response = await self.client.search(
            index=self._index_name_for_dataset(dataset),
            body=body,
            from_=offset,
            size=limit,
            _source=False,
            sort="_score:desc,id:asc",
            track_total_hits=True,
        )

        items = [
            SearchResponseItem(record_id=UUID(hit["_id"]), score=hit["_score"]) for hit in response["hits"]["hits"]
        ]
        total = response["hits"]["total"]["value"]

        return SearchResponses(items=items, total=total)

    @staticmethod
    def _text_query_builder(dataset: Dataset, text: TextQuery) -> dict:
        if not text.field:
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-multi-match-query.html
            field_names = [
                f"fields.{field.name}" for field in dataset.fields if field.settings.get("type") == FieldType.text
            ]
            return {"multi_match": {"query": text.q, "fields": field_names, "operator": "and"}}
        else:
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-match-query.html
            return {"match": {f"fields.{text.field}": {"query": text.q, "operator": "and"}}}

    def _mapping_for_fields(self, fields: List[Field]):
        return {f"fields.{field.name}": self._es_mapping_for_field(field) for field in fields}

    def _dynamic_templates_for_question_responses(self, questions: List[Question]) -> List[dict]:
        # See https://www.elastic.co/guide/en/elasticsearch/reference/current/dynamic-templates.html
        return [
            {"status_responses": {"path_match": f"responses.*.status", "mapping": {"type": "keyword"}}},
            *[
                {
                    f"{question.name}_responses": {
                        "path_match": f"responses.*.values.{question.name}",
                        "mapping": self._field_mapping_for_question(question),
                    },
                }
                for question in questions
            ],
        ]

    def _field_mapping_for_question(self, question: Question):
        settings = question.parsed_settings

        if settings.type == QuestionType.rating:
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/number.html
            return {"type": "integer"}
        elif settings.type == QuestionType.text:
            # TODO: Review mapping for label selection. Could make sense to use `keyword` mapping instead.
            #  See https://www.elastic.co/guide/en/elasticsearch/reference/current/keyword.html
            #  See https://www.elastic.co/guide/en/elasticsearch/reference/current/text.html
            return {"type": "text", "index": False}
        elif settings.type in [QuestionType.label_selection, QuestionType.multi_label_selection]:
            return {"type": "keyword"}
        elif settings.type == QuestionType.ranking:
            return {"type": "nested"}
        else:
            raise ValueError(f"ElasticSearch mappings for Question of type {settings.type} cannot be generated")

    @staticmethod
    def _es_mapping_for_field(field: Field):
        field_type = field.settings["type"]

        if field_type == FieldType.text:
            return {"type": "text"}
        else:
            raise ValueError(f"ElasticSearch mappings for Field of type {field_type} cannot be generated")

    async def _get_index_or_raise(self, dataset: Dataset):
        index_name = self._index_name_for_dataset(dataset)
        if not await self.client.indices.exists(index=index_name):
            raise ValueError(f"Cannot access to index for dataset {dataset.id}: the specified index does not exist")

        return index_name

    @staticmethod
    def _index_name_for_dataset(dataset: Dataset):
        return f"rg.{dataset.id}"

    def _response_status_filter_builder(self, status_filter: UserResponseStatusFilter):
        user_response_field = f"responses.{status_filter.user.username}"

        if status_filter.status == ResponseStatusFilter.missing:
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-exists-query.html
            return [{"bool": {"must_not": {"exists": {"field": user_response_field}}}}]
        else:
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-terms-query.html
            return [{"term": {f"{user_response_field}.status": status_filter.status}}]


async def get_search_engine() -> AsyncGenerator[SearchEngine, None]:
    config = dict(
        hosts=settings.elasticsearch,
        verify_certs=settings.elasticsearch_ssl_verify,
        ca_certs=settings.elasticsearch_ca_path,
        retry_on_timeout=True,
        max_retries=5,
    )
    search_engine = SearchEngine(
        config,
        es_number_of_shards=settings.es_records_index_shards,
        es_number_of_replicas=settings.es_records_index_shards,
    )
    try:
        yield search_engine
    finally:
        await search_engine.client.close()



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List, Optional, Union

import httpx

from argilla.client.sdk._helpers import build_typed_response
from argilla.client.sdk.client import AuthenticatedClient
from argilla.client.sdk.commons.api import (
    build_data_response,
    build_list_response,
    build_param_dict,
)
from argilla.client.sdk.commons.models import (
    ErrorMessage,
    HTTPValidationError,
    Response,
)
from argilla.client.sdk.text_classification.models import (
    LabelingRule,
    LabelingRuleMetricsSummary,
    TextClassificationQuery,
    TextClassificationRecord,
)


def add_dataset_labeling_rule(
    client: AuthenticatedClient,
    name: str,
    rule: LabelingRule,
) -> Response[Union[LabelingRule, HTTPValidationError, ErrorMessage]]:
    url = "{}/api/datasets/{name}/TextClassification/labeling/rules".format(client.base_url, name=name)

    response = httpx.post(
        url=url,
        json={"query": rule.query, "labels": rule.labels},
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
    )

    return build_typed_response(response, LabelingRule)


def update_dataset_labeling_rule(
    client: AuthenticatedClient,
    name: str,
    rule: LabelingRule,
) -> Response[Union[HTTPValidationError, ErrorMessage]]:
    url = "{}/api/datasets/TextClassification/{name}/labeling/rules/{query}".format(
        client.base_url, name=name, query=rule.query
    )

    response = httpx.patch(
        url,
        json={"labels": rule.labels},
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
    )

    return build_typed_response(response, LabelingRule)


def delete_dataset_labeling_rule(
    client: AuthenticatedClient,
    name: str,
    rule: LabelingRule,
) -> Response[Union[LabelingRule, HTTPValidationError, ErrorMessage]]:
    url = "{}/api/datasets/TextClassification/{name}/labeling/rules/{query}".format(
        client.base_url, name=name, query=rule.query
    )

    httpx.delete(
        url,
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
    )


def fetch_dataset_labeling_rules(
    client: AuthenticatedClient,
    name: str,
) -> Response[Union[List[LabelingRule], HTTPValidationError, ErrorMessage]]:
    url = "{}/api/datasets/TextClassification/{name}/labeling/rules".format(client.base_url, name=name)

    response = httpx.get(
        url=url,
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
    )

    return build_list_response(response, LabelingRule)


def dataset_rule_metrics(
    client: AuthenticatedClient,
    name: str,
    query: str,
    label: str,
) -> Response[Union[LabelingRuleMetricsSummary, HTTPValidationError, ErrorMessage]]:
    url = "{}/api/datasets/TextClassification/{name}/labeling/rules/{query}/metrics?label={label}".format(
        client.base_url, name=name, query=query, label=label
    )

    response = httpx.get(
        url=url,
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
    )

    return build_typed_response(response, response_type_class=LabelingRuleMetricsSummary)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import Any, Dict

from fastapi import HTTPException, Request
from fastapi.exception_handlers import http_exception_handler
from pydantic import BaseModel

from argilla.server.commons import telemetry
from argilla.server.errors.adapter import exception_to_argilla_error
from argilla.server.errors.base_errors import ServerError

_LOGGER = logging.getLogger("argilla")


class ErrorDetail(BaseModel):
    code: str
    params: Dict[str, Any]


# TODO(@frascuchon): Review class Naming
class ServerHTTPException(HTTPException):
    def __init__(self, error: ServerError):
        super().__init__(
            status_code=error.HTTP_STATUS,
            detail=ErrorDetail(code=error.code, params=error.arguments).dict(),
        )


class APIErrorHandler:
    @staticmethod
    async def common_exception_handler(request: Request, error: Exception):
        """Wraps errors as custom generic error"""
        argilla_error = exception_to_argilla_error(error)
        await telemetry.track_error(argilla_error, request=request)

        return await http_exception_handler(request, ServerHTTPException(argilla_error))



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""add record metadata column

Revision ID: 3ff6484f8b37
Revises: ae5522b4c674
Create Date: 2023-06-14 13:02:41.735153

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3ff6484f8b37"
down_revision = "ae5522b4c674"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("records", sa.Column("metadata", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("records", "metadata")
    # ### end Alembic commands ###



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from types import TracebackType
from typing import ContextManager, Optional, Type

from packaging.version import parse
from pydantic import BaseModel

from argilla.client.apis import AbstractApi
from argilla.client.sdk.client import AuthenticatedClient
from argilla.client.sdk.commons.errors import ApiCompatibilityError


@dataclasses.dataclass(frozen=True)
class ApiInfo:
    version: Optional[str] = None


class Status(AbstractApi):
    class _ApiInfo(BaseModel):
        version: Optional[str] = None

    def get_info(self) -> ApiInfo:
        """
        Get the connected API info

        """

        response = self.http_client.get("/api/_info")
        api_info = self._ApiInfo.parse_obj(response)
        return ApiInfo(version=api_info.version)


class _ApiCompatibilityContextManager(ContextManager):
    def __init__(
        self,
        client: AuthenticatedClient,
        min_version: str,
    ):
        self._min_version = parse(min_version)
        self._status_api = Status(client=client) if client else None

    def __enter__(self):
        if self._status_api:
            api_info = self._status_api.get_info()
            api_version = api_info.version

            api_version = parse(api_version)
            if api_version.is_devrelease:
                api_version = parse(api_version.base_version)
            if not api_version >= self._min_version:
                raise ApiCompatibilityError(str(self._min_version), api_version=api_version)
        pass

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        pass


def api_compatibility(api: AbstractApi, min_version: str):
    """
    Handles problems related to server API compatibility.

    Args:
        api: The api component where apply the api compatibility
        min_version: the minimal version that the argilla client can work with
    """
    return _ApiCompatibilityContextManager(
        min_version=min_version,
        client=api.http_client,
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from enum import Enum
from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel, Field


class UserRole(str, Enum):
    owner = "owner"
    admin = "admin"
    annotator = "annotator"


class UserCreateModel(BaseModel):
    first_name: str = Field(min_length=1)
    last_name: Optional[str] = Field(min_length=1)
    username: str = Field(min_length=1, regex=r"^(?!-|_)[a-z0-9-_]+$")
    role: UserRole = UserRole.annotator
    password: str = Field(min_length=8, max_length=100)

    # TODO(alvarobartt): confirm with @frascuchon
    # disabled: Optional[bool] = None
    # api_key: Optional[str]  # backward compatibility


class UserModel(BaseModel):
    id: UUID
    first_name: str
    last_name: Optional[str]
    full_name: Optional[str]
    username: str
    role: UserRole
    workspaces: Optional[List[str]]
    api_key: str
    inserted_at: datetime
    updated_at: datetime



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, PositiveInt, conlist, constr, root_validator, validator
from pydantic import Field as PydanticField
from pydantic.utils import GetterDict

from argilla.server.search_engine import Query

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

from argilla.server.models import (
    DatasetStatus,
    FieldType,
    QuestionSettings,
    QuestionType,
    ResponseStatus,
)

DATASET_CREATE_GUIDELINES_MIN_LENGTH = 1
DATASET_CREATE_GUIDELINES_MAX_LENGTH = 10000

FIELD_CREATE_NAME_REGEX = r"^(?=.*[a-z0-9])[a-z0-9_-]+$"
FIELD_CREATE_NAME_MIN_LENGTH = 1
FIELD_CREATE_NAME_MAX_LENGTH = 200
FIELD_CREATE_TITLE_MIN_LENGTH = 1
FIELD_CREATE_TITLE_MAX_LENGTH = 500

QUESTION_CREATE_NAME_REGEX = r"^(?=.*[a-z0-9])[a-z0-9_-]+$"
QUESTION_CREATE_NAME_MIN_LENGTH = 1
QUESTION_CREATE_NAME_MAX_LENGTH = 200
QUESTION_CREATE_TITLE_MIN_LENGTH = 1
QUESTION_CREATE_TITLE_MAX_LENGTH = 500
QUESTION_CREATE_DESCRIPTION_MIN_LENGTH = 1
QUESTION_CREATE_DESCRIPTION_MAX_LENGTH = 1000

RATING_OPTIONS_MIN_ITEMS = 2
RATING_OPTIONS_MAX_ITEMS = 100

VALUE_TEXT_OPTION_VALUE_MIN_LENGHT = 1
VALUE_TEXT_OPTION_VALUE_MAX_LENGTH = 200
VALUE_TEXT_OPTION_TEXT_MIN_LENGTH = 1
VALUE_TEXT_OPTION_TEXT_MAX_LENGTH = 500
VALUE_TEXT_OPTION_DESCRIPTION_MIN_LENGTH = 1
VALUE_TEXT_OPTION_DESCRIPTION_MAX_LENGTH = 1000

LABEL_SELECTION_OPTIONS_MIN_ITEMS = 2
LABEL_SELECTION_OPTIONS_MAX_ITEMS = 250

RANKING_OPTIONS_MIN_ITEMS = 2


RECORDS_CREATE_MIN_ITEMS = 1
RECORDS_CREATE_MAX_ITEMS = 1000


class Dataset(BaseModel):
    id: UUID
    name: str
    guidelines: Optional[str]
    status: DatasetStatus
    workspace_id: UUID
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class Datasets(BaseModel):
    items: List[Dataset]


class DatasetCreate(BaseModel):
    name: str
    guidelines: Optional[
        constr(
            min_length=DATASET_CREATE_GUIDELINES_MIN_LENGTH,
            max_length=DATASET_CREATE_GUIDELINES_MAX_LENGTH,
        )
    ]
    workspace_id: UUID


class RecordMetrics(BaseModel):
    count: int


class ResponseMetrics(BaseModel):
    count: int
    submitted: int
    discarded: int
    draft: int


class Metrics(BaseModel):
    records: RecordMetrics
    responses: ResponseMetrics


class TextFieldSettings(BaseModel):
    type: Literal[FieldType.text]
    use_markdown: bool = False


class Field(BaseModel):
    id: UUID
    name: str
    title: str
    required: bool
    settings: TextFieldSettings
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class Fields(BaseModel):
    items: List[Field]


class FieldCreate(BaseModel):
    name: constr(
        regex=FIELD_CREATE_NAME_REGEX,
        min_length=FIELD_CREATE_NAME_MIN_LENGTH,
        max_length=FIELD_CREATE_NAME_MAX_LENGTH,
    )
    title: constr(
        min_length=FIELD_CREATE_TITLE_MIN_LENGTH,
        max_length=FIELD_CREATE_TITLE_MAX_LENGTH,
    )
    required: Optional[bool]
    settings: TextFieldSettings


class TextQuestionSettingsCreate(BaseModel):
    type: Literal[QuestionType.text]
    use_markdown: bool = False


class UniqueValuesCheckerMixin(BaseModel):
    @root_validator
    def check_unique_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        options = values.get("options", [])
        seen = set()
        duplicates = set()
        for option in options:
            if option.value in seen:
                duplicates.add(option.value)
            else:
                seen.add(option.value)
        if duplicates:
            raise ValueError(f"Option values must be unique, found duplicates: {duplicates}")
        return values


class RatingQuestionSettingsOption(BaseModel):
    value: int


class RatingQuestionSettingsCreate(UniqueValuesCheckerMixin):
    type: Literal[QuestionType.rating]
    options: conlist(
        item_type=RatingQuestionSettingsOption,
        min_items=RATING_OPTIONS_MIN_ITEMS,
        max_items=RATING_OPTIONS_MAX_ITEMS,
    )


class ValueTextQuestionSettingsOption(BaseModel):
    value: constr(
        min_length=VALUE_TEXT_OPTION_VALUE_MIN_LENGHT,
        max_length=VALUE_TEXT_OPTION_VALUE_MAX_LENGTH,
    )
    text: constr(
        min_length=VALUE_TEXT_OPTION_TEXT_MIN_LENGTH,
        max_length=VALUE_TEXT_OPTION_TEXT_MAX_LENGTH,
    )
    description: Optional[
        constr(
            min_length=VALUE_TEXT_OPTION_DESCRIPTION_MIN_LENGTH,
            max_length=VALUE_TEXT_OPTION_DESCRIPTION_MAX_LENGTH,
        )
    ] = None


class LabelSelectionQuestionSettingsCreate(UniqueValuesCheckerMixin):
    type: Literal[QuestionType.label_selection]
    options: conlist(
        item_type=ValueTextQuestionSettingsOption,
        min_items=LABEL_SELECTION_OPTIONS_MIN_ITEMS,
        max_items=LABEL_SELECTION_OPTIONS_MAX_ITEMS,
    )
    visible_options: Optional[PositiveInt] = None


class MultiLabelSelectionQuestionSettingsCreate(LabelSelectionQuestionSettingsCreate):
    type: Literal[QuestionType.multi_label_selection]


class RankingQuestionSettingsCreate(UniqueValuesCheckerMixin):
    type: Literal[QuestionType.ranking]
    options: conlist(
        item_type=ValueTextQuestionSettingsOption,
        min_items=RANKING_OPTIONS_MIN_ITEMS,
    )


QuestionSettingsCreate = Annotated[
    Union[
        TextQuestionSettingsCreate,
        RatingQuestionSettingsCreate,
        LabelSelectionQuestionSettingsCreate,
        MultiLabelSelectionQuestionSettingsCreate,
        RankingQuestionSettingsCreate,
    ],
    PydanticField(discriminator="type"),
]


class Question(BaseModel):
    id: UUID
    name: str
    title: str
    description: Optional[str]
    required: bool
    settings: QuestionSettings
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class Questions(BaseModel):
    items: List[Question]


class QuestionCreate(BaseModel):
    name: constr(
        regex=QUESTION_CREATE_NAME_REGEX,
        min_length=QUESTION_CREATE_NAME_MIN_LENGTH,
        max_length=QUESTION_CREATE_NAME_MAX_LENGTH,
    )
    title: constr(
        min_length=QUESTION_CREATE_TITLE_MIN_LENGTH,
        max_length=QUESTION_CREATE_TITLE_MAX_LENGTH,
    )
    description: Optional[
        constr(
            min_length=QUESTION_CREATE_DESCRIPTION_MIN_LENGTH,
            max_length=QUESTION_CREATE_DESCRIPTION_MAX_LENGTH,
        )
    ]
    required: Optional[bool]
    settings: QuestionSettingsCreate


class ResponseValue(BaseModel):
    value: Any


class ResponseValueCreate(BaseModel):
    value: Any


class Response(BaseModel):
    id: UUID
    values: Optional[Dict[str, ResponseValue]]
    status: ResponseStatus
    user_id: UUID
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class RecordInclude(str, Enum):
    responses = "responses"


class RecordGetterDict(GetterDict):
    def get(self, key: str, default: Any) -> Any:
        if key == "metadata":
            return getattr(self._obj, "metadata_", None)
        if key == "responses" and "responses" not in self._obj.__dict__:
            return default
        return super().get(key, default)


class Record(BaseModel):
    id: UUID
    fields: Dict[str, Any]
    metadata: Optional[Dict[str, Any]]
    external_id: Optional[str]
    # TODO: move `responses` to `response` since contextualized endpoint will contains only the user response
    # response: Optional[Response]
    responses: Optional[List[Response]]
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True
        getter_dict = RecordGetterDict


class Records(BaseModel):
    items: List[Record]


class UserSubmittedResponseCreate(BaseModel):
    user_id: UUID
    values: Dict[str, ResponseValueCreate]
    status: Literal[ResponseStatus.submitted]


class UserDiscardedResponseCreate(BaseModel):
    user_id: UUID
    values: Optional[Dict[str, ResponseValueCreate]]
    status: Literal[ResponseStatus.discarded]


UserResponseCreate = Annotated[
    Union[UserSubmittedResponseCreate, UserDiscardedResponseCreate],
    PydanticField(discriminator="status"),
]


class RecordCreate(BaseModel):
    fields: Dict[str, Any]
    metadata: Optional[Dict[str, Any]]
    external_id: Optional[str]
    responses: Optional[List[UserResponseCreate]]

    @validator("responses")
    def check_user_id_is_unique(cls, values):
        user_ids = []

        for value in values:
            if value.user_id in user_ids:
                raise ValueError(f"Responses contains several responses for the same user_id: {str(value.user_id)!r}")
            user_ids.append(value.user_id)

        return values


class RecordsCreate(BaseModel):
    items: conlist(item_type=RecordCreate, min_items=RECORDS_CREATE_MIN_ITEMS, max_items=RECORDS_CREATE_MAX_ITEMS)


class SearchRecordsQuery(BaseModel):
    query: Query


class SearchRecord(BaseModel):
    record: Record
    query_score: Optional[float]


class SearchRecordsResult(BaseModel):
    items: List[SearchRecord]
    total: int = 0



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Query, Security, status
from sqlalchemy.orm import Session

from argilla.server.commons.telemetry import TelemetryClient, get_telemetry_client
from argilla.server.contexts import accounts, datasets
from argilla.server.database import get_db
from argilla.server.enums import ResponseStatusFilter
from argilla.server.models import ResponseStatus, User
from argilla.server.policies import DatasetPolicyV1, authorize
from argilla.server.schemas.v1.datasets import (
    Dataset,
    DatasetCreate,
    Datasets,
    Field,
    FieldCreate,
    Fields,
    Metrics,
    Question,
    QuestionCreate,
    Questions,
    Record,
    RecordInclude,
    Records,
    RecordsCreate,
    SearchRecord,
    SearchRecordsQuery,
    SearchRecordsResult,
)
from argilla.server.search_engine import (
    SearchEngine,
    UserResponseStatusFilter,
    get_search_engine,
)
from argilla.server.security import auth

LIST_DATASET_RECORDS_LIMIT_DEFAULT = 50
LIST_DATASET_RECORDS_LIMIT_LTE = 1000

router = APIRouter(tags=["datasets"])


def _get_dataset(db: Session, dataset_id: UUID):
    dataset = datasets.get_dataset_by_id(db, dataset_id)
    if not dataset:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Dataset with id `{dataset_id}` not found",
        )

    return dataset


@router.get("/me/datasets", response_model=Datasets)
def list_current_user_datasets(
    *,
    db: Session = Depends(get_db),
    current_user: User = Security(auth.get_current_user),
):
    authorize(current_user, DatasetPolicyV1.list)

    if current_user.is_owner:
        return Datasets(items=datasets.list_datasets(db))
    else:
        return Datasets(items=current_user.datasets)


@router.get("/datasets/{dataset_id}/fields", response_model=Fields)
def list_dataset_fields(
    *,
    db: Session = Depends(get_db),
    dataset_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.get(dataset))

    return Fields(items=dataset.fields)


@router.get("/datasets/{dataset_id}/questions", response_model=Questions)
def list_dataset_questions(
    *,
    db: Session = Depends(get_db),
    dataset_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.get(dataset))

    return Questions(items=dataset.questions)


@router.get("/me/datasets/{dataset_id}/records", response_model=Records, response_model_exclude_unset=True)
def list_current_user_dataset_records(
    *,
    db: Session = Depends(get_db),
    dataset_id: UUID,
    include: List[RecordInclude] = Query([]),
    response_status: Optional[ResponseStatusFilter] = Query(None),
    offset: int = 0,
    limit: int = Query(default=LIST_DATASET_RECORDS_LIMIT_DEFAULT, lte=LIST_DATASET_RECORDS_LIMIT_LTE),
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.get(dataset))

    records = datasets.list_records_by_dataset_id_and_user_id(
        db, dataset_id, current_user.id, include=include, response_status=response_status, offset=offset, limit=limit
    )

    return Records(items=records)


@router.get("/datasets/{dataset_id}/records", response_model=Records, response_model_exclude_unset=True)
def list_dataset_records(
    *,
    db: Session = Depends(get_db),
    dataset_id: UUID,
    include: Optional[List[RecordInclude]] = Query([]),
    offset: int = 0,
    limit: int = Query(default=LIST_DATASET_RECORDS_LIMIT_DEFAULT, lte=LIST_DATASET_RECORDS_LIMIT_LTE),
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.list_dataset_records_will_all_responses(dataset))

    records = datasets.list_records_by_dataset_id(db, dataset_id, include=include, offset=offset, limit=limit)

    return Records(items=records)


@router.get("/datasets/{dataset_id}", response_model=Dataset)
def get_dataset(
    *,
    db: Session = Depends(get_db),
    dataset_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.get(dataset))

    return dataset


@router.get("/me/datasets/{dataset_id}/metrics", response_model=Metrics)
def get_current_user_dataset_metrics(
    *,
    db: Session = Depends(get_db),
    dataset_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.get(dataset))

    return {
        "records": {
            "count": datasets.count_records_by_dataset_id(db, dataset_id),
        },
        "responses": {
            "count": datasets.count_responses_by_dataset_id_and_user_id(db, dataset_id, current_user.id),
            "submitted": datasets.count_responses_by_dataset_id_and_user_id(
                db, dataset_id, current_user.id, ResponseStatus(ResponseStatusFilter.submitted)
            ),
            "discarded": datasets.count_responses_by_dataset_id_and_user_id(
                db, dataset_id, current_user.id, ResponseStatus(ResponseStatusFilter.discarded)
            ),
            "draft": datasets.count_responses_by_dataset_id_and_user_id(
                db, dataset_id, current_user.id, ResponseStatus(ResponseStatusFilter.draft)
            ),
        },
    }


@router.post("/datasets", status_code=status.HTTP_201_CREATED, response_model=Dataset)
def create_dataset(
    *,
    db: Session = Depends(get_db),
    dataset_create: DatasetCreate,
    current_user: User = Security(auth.get_current_user),
):
    authorize(current_user, DatasetPolicyV1.create(dataset_create.workspace_id))

    if not accounts.get_workspace_by_id(db, dataset_create.workspace_id):
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Workspace with id `{dataset_create.workspace_id}` not found",
        )

    if datasets.get_dataset_by_name_and_workspace_id(db, dataset_create.name, dataset_create.workspace_id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Dataset with name `{dataset_create.name}` already exists for workspace with id `{dataset_create.workspace_id}`",
        )

    return datasets.create_dataset(db, dataset_create)


@router.post("/datasets/{dataset_id}/fields", status_code=status.HTTP_201_CREATED, response_model=Field)
def create_dataset_field(
    *,
    db: Session = Depends(get_db),
    dataset_id: UUID,
    field_create: FieldCreate,
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.create_field(dataset))

    if datasets.get_field_by_name_and_dataset_id(db, field_create.name, dataset_id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Field with name `{field_create.name}` already exists for dataset with id `{dataset_id}`",
        )

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        return datasets.create_field(db, dataset, field_create)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.post("/datasets/{dataset_id}/questions", status_code=status.HTTP_201_CREATED, response_model=Question)
def create_dataset_question(
    *,
    db: Session = Depends(get_db),
    dataset_id: UUID,
    question_create: QuestionCreate,
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.create_question(dataset))

    if datasets.get_question_by_name_and_dataset_id(db, question_create.name, dataset_id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Question with name `{question_create.name}` already exists for dataset with id `{dataset_id}`",
        )

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        return datasets.create_question(db, dataset, question_create)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.post("/datasets/{dataset_id}/records", status_code=status.HTTP_204_NO_CONTENT)
async def create_dataset_records(
    *,
    db: Session = Depends(get_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    telemetry_client: TelemetryClient = Depends(get_telemetry_client),
    dataset_id: UUID,
    records_create: RecordsCreate,
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.create_records(dataset))

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    #  After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        await datasets.create_records(db, search_engine, dataset=dataset, records_create=records_create)
        telemetry_client.track_data(action="DatasetRecordsCreated", data={"records": len(records_create.items)})
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.post(
    "/me/datasets/{dataset_id}/records/search",
    status_code=status.HTTP_200_OK,
    response_model=SearchRecordsResult,
    response_model_exclude_unset=True,
)
async def search_dataset_records(
    *,
    db: Session = Depends(get_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    telemetry_client: TelemetryClient = Depends(get_telemetry_client),
    dataset_id: UUID,
    query: SearchRecordsQuery,
    include: List[RecordInclude] = Query([]),
    response_status: Optional[ResponseStatusFilter] = Query(None),
    offset: int = Query(0, ge=0),
    limit: int = Query(default=LIST_DATASET_RECORDS_LIMIT_DEFAULT, lte=LIST_DATASET_RECORDS_LIMIT_LTE),
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)
    authorize(current_user, DatasetPolicyV1.search_records(dataset))

    search_engine_query = query.query
    if search_engine_query.text.field and not datasets.get_field_by_name_and_dataset_id(
        db, search_engine_query.text.field, dataset_id
    ):
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Field `{search_engine_query.text.field}` not found in dataset `{dataset_id}`.",
        )

    user_response_status_filter = None
    if response_status:
        user_response_status_filter = UserResponseStatusFilter(user=current_user, status=response_status)

    search_responses = await search_engine.search(
        dataset=dataset,
        query=search_engine_query,
        user_response_status_filter=user_response_status_filter,
        offset=offset,
        limit=limit,
    )

    record_id_score_map = {
        response.record_id: {"query_score": response.score, "search_record": None}
        for response in search_responses.items
    }
    records = datasets.get_records_by_ids(
        db=db,
        dataset_id=dataset_id,
        record_ids=list(record_id_score_map.keys()),
        include=include,
        user_id=current_user.id,
    )

    for record in records:
        record_id_score_map[record.id]["search_record"] = SearchRecord(
            record=Record.from_orm(record), query_score=record_id_score_map[record.id]["query_score"]
        )

    return SearchRecordsResult(
        items=[record["search_record"] for record in record_id_score_map.values()], total=search_responses.total
    )


@router.put("/datasets/{dataset_id}/publish", response_model=Dataset)
async def publish_dataset(
    *,
    db: Session = Depends(get_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    telemetry_client: TelemetryClient = Depends(get_telemetry_client),
    dataset_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.publish(dataset))
    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    #  After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        dataset = await datasets.publish_dataset(db, search_engine, dataset)

        telemetry_client.track_data(
            action="PublishedDataset",
            data={"questions": list(set([question.settings["type"] for question in dataset.questions]))},
        )

        return dataset
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.delete("/datasets/{dataset_id}", response_model=Dataset)
async def delete_dataset(
    *,
    db: Session = Depends(get_db),
    search_engine=Depends(get_search_engine),
    dataset_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    dataset = _get_dataset(db, dataset_id)

    authorize(current_user, DatasetPolicyV1.delete(dataset))

    await datasets.delete_dataset(db, search_engine, dataset=dataset)

    return dataset



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Any

import httpx


class BaseClientError(Exception):
    pass


class WrongResponseError(BaseClientError):
    def __init__(self, message: str, response: Any):
        self.message = message
        self.response = response

    def __str__(self):
        return f"\nUnexpected response: {self.message}" "\nResponse content:" f"\n{self.response}"


class InputValueError(BaseClientError):
    pass


class ApiCompatibilityError(BaseClientError):
    def __init__(self, min_version: str, api_version: str):
        self.min_version = min_version
        self.api_version = api_version

    def __str__(self):
        return (
            "\nThe argilla server does not support this functionality."
            f"\nPlease, use the client with a {self.min_version} version server instance."
        )


class HttpResponseError(BaseClientError):
    """Used for handle http errros other than defined in Argilla server"""

    def __init__(self, response: httpx.Response):
        self.status_code = response.status_code
        self.detail = response.text

    def __str__(self):
        return (
            "Received an HTTP error from server\n"
            + f"Response status: {self.status_code}\n"
            + f"Response content: {self.detail}\n"
        )


class ArApiResponseError(BaseClientError):
    HTTP_STATUS: int

    def __init__(self, **ctx):
        self.ctx = ctx

    def __str__(self):
        return (
            f"Argilla server returned an error with http status: {self.HTTP_STATUS}" + f"\nError details: [{self.ctx}]"
        )


class BadRequestApiError(ArApiResponseError):
    HTTP_STATUS = 400


class UnauthorizedApiError(ArApiResponseError):
    HTTP_STATUS = 401


class ForbiddenApiError(ArApiResponseError):
    HTTP_STATUS = 403


class NotFoundApiError(ArApiResponseError):
    HTTP_STATUS = 404


class MethodNotAllowedApiError(ArApiResponseError):
    HTTP_STATUS = 405


class AlreadyExistsApiError(ArApiResponseError):
    HTTP_STATUS = 409


class ValidationApiError(ArApiResponseError):
    HTTP_STATUS = 422

    def __init__(self, client_ctx, params, **ctx):
        for error in params.get("errors", []):
            current_level = client_ctx
            for loc in error["loc"]:
                new_value = None
                try:
                    new_value = current_level[loc]
                except KeyError:
                    pass
                if new_value is None:
                    break
                if hasattr(new_value, "dict"):
                    new_value = new_value.dict()
                current_level = new_value
            error["value"] = current_level

        # TODO: parse error details and match with client context
        super().__init__(**ctx, params=params)


class GenericApiError(ArApiResponseError):
    HTTP_STATUS = 500



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import inspect
from typing import Dict

args_to_remove = ["self", "cls", "model_args", "args", "kwargs", "model_kwargs"]


def get_default_args(fn) -> dict:
    """
    It takes a function and returns a dictionary of the default arguments

    Args:
      func: The function you want to get the default arguments from.

    Returns:
      A dictionary of the default arguments for the function.
    """
    arg_spec = inspect.getfullargspec(fn)
    # `arg_spec` contains the default values for the optional arguments, starting on the first optional, so there
    # may be mandatory arguments before with no default value, those will be set to `None`, while the rest should match
    # the default values in `arg_spec.defaults`
    default_args = (
        dict(zip(arg_spec.args, [None] * (len(arg_spec.args) - len(arg_spec.defaults)) + list(arg_spec.defaults)))
        if arg_spec.defaults
        else {}
    )
    for key in args_to_remove:
        if key in default_args:
            del default_args[key]
    return default_args


def filter_allowed_args(
    func,
    **kwargs,
):
    """
    It takes a function and a dictionary of arguments, and returns a dictionary of arguments that are
    allowed by the function

    Args:
      func: The function to filter the arguments for.

    Returns:
      A dictionary of the arguments that are allowed in the function.
    """
    allowed_args = {key: val for key, val in kwargs.items() if key in func.__code__.co_varnames}
    for key in args_to_remove:
        if key in allowed_args:
            del allowed_args[key]
    return allowed_args


def _apply_column_mapping(dataset: "Dataset", column_mapping: Dict[str, str]) -> "Dataset":
    """
    Applies the provided column mapping to the dataset, renaming columns accordingly.
    Extra features not in the column mapping are prefixed with `"feat_"`.
    """
    column_mapping = {label_from: label_to for label_to, label_from in column_mapping.items()}
    dataset = dataset.rename_columns(
        {
            **column_mapping,
            **{col: f"feat_{col}" for col in dataset.column_names if col not in column_mapping},
        }
    )
    dset_format = dataset.format
    dataset = dataset.with_format(
        type=dset_format["type"],
        columns=dataset.column_names,
        output_all_columns=dset_format["output_all_columns"],
        **dset_format["format_kwargs"],
    )
    return dataset



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, ClassVar, Dict, Iterable, List

from pydantic import Field

from argilla.server.services.metrics import ServiceBaseMetric, ServicePythonMetric
from argilla.server.services.metrics.models import CommonTasksMetrics
from argilla.server.services.search.model import ServiceRecordsQuery
from argilla.server.services.tasks.text_classification.model import (
    ServiceTextClassificationRecord,
)
from argilla.utils.dependency import requires_version


class F1Metric(ServicePythonMetric):
    """
    A basic f1 computation for text classification

    Attributes:
    -----------
        multi_label:
            If True, F1 will be calculated assuming multi class task. Default False
    """

    multi_label: bool = False

    @requires_version("scikit-learn")
    def apply(self, records: Iterable[ServiceTextClassificationRecord]) -> Any:
        from sklearn.metrics import precision_recall_fscore_support

        filtered_records = list(filter(lambda r: r.predicted is not None, records))
        # TODO: This must be precomputed with using a global dataset metric
        ds_labels = {label for record in filtered_records for label in record.annotated_as + record.predicted_as}

        if not len(ds_labels):
            return {}

        labels_mapping = {label: i for i, label in enumerate(ds_labels)}
        y_true, y_pred = ([], [])
        for record in filtered_records:
            annotations = record.annotated_as
            predictions = record.predicted_as

            if not self.multi_label:
                y_true.append(labels_mapping[annotations[0]])
                y_pred.append(labels_mapping[predictions[0]])

            else:
                y_true.append([labels_mapping[label] for label in annotations])
                y_pred.append([labels_mapping[label] for label in predictions])

        if self.multi_label:
            from sklearn.preprocessing import MultiLabelBinarizer

            mlb = MultiLabelBinarizer(classes=list(labels_mapping.values()))
            y_true = mlb.fit_transform(y_true)
            y_pred = mlb.fit_transform(y_pred)

        micro_p, micro_r, micro_f, _ = precision_recall_fscore_support(y_true=y_true, y_pred=y_pred, average="micro")
        macro_p, macro_r, macro_f, _ = precision_recall_fscore_support(y_true=y_true, y_pred=y_pred, average="macro")

        per_label = {}
        for label, p, r, f, s in zip(
            labels_mapping.keys(),
            *precision_recall_fscore_support(
                y_true=y_true,
                y_pred=y_pred,
                average=None,
            ),
        ):
            per_label.update(
                {
                    f"{label}_precision": p.tolist(),
                    f"{label}_recall": r.tolist(),
                    f"{label}_f1": f.tolist(),
                    f"{label}_support": s.tolist(),
                }
            )

        return {
            "precision_macro": macro_p,
            "recall_macro": macro_r,
            "f1_macro": macro_f,
            "precision_micro": micro_p,
            "recall_micro": micro_r,
            "f1_micro": micro_f,
            **per_label,
        }


class DatasetLabels(ServicePythonMetric):
    id: str = Field("dataset_labels", const=True)
    name: str = Field("The dataset labels", const=True)

    records_to_fetch: int = 1000

    shuffle_records: bool = Field(default=True)

    def prepare_query(self, query: ServiceRecordsQuery):
        text = query.query_text or ""
        if text:
            text += " AND "
        text += "(_exists_:annotated_by OR _exists_:predicted_by)"

        query.query_text = text
        return query

    def apply(
        self,
        records: Iterable[ServiceTextClassificationRecord],
    ) -> Dict[str, Any]:
        ds_labels = set()
        for _ in range(0, self.records_to_fetch):  # Only a few of records will be parsed
            record = next(records, None)
            if record is None:
                break

            if record.annotation:
                ds_labels.update([label.class_label for label in record.annotation.labels])
            if record.prediction:
                ds_labels.update([label.class_label for label in record.prediction.labels])
        return {"labels": ds_labels or []}


class TextClassificationMetrics(CommonTasksMetrics[ServiceTextClassificationRecord]):
    """Configured metrics for text classification task"""

    metrics: ClassVar[List[ServiceBaseMetric]] = (
        CommonTasksMetrics.metrics
        + [
            F1Metric(
                id="F1",
                name="F1 Metrics for single-label",
                description="F1 Metrics for single-label (averaged and per label)",
            ),
            F1Metric(
                id="MultiLabelF1",
                name="F1 Metrics for multi-label",
                description="F1 Metrics for multi-label (averaged and per label)",
                multi_label=True,
            ),
            DatasetLabels(),
        ]
        + [
            ServiceBaseMetric(
                id="predicted_as",
                name="Predicted labels distribution",
            ),
            ServiceBaseMetric(
                id="annotated_as",
                name="Annotated labels distribution",
            ),
            ServiceBaseMetric(
                id="labeling_rule",
                name="Labeling rule metric based on a query rule and a set of labels",
            ),
            ServiceBaseMetric(
                id="dataset_labeling_rules",
                name="Computes the overall labeling rules stats",
            ),
        ]
    )



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Common model for task definitions
"""

from typing import Any, Dict, Generic, List, TypeVar

from pydantic import BaseModel, Field
from pydantic.generics import GenericModel

from argilla.server.services.search.model import (
    ServiceQueryRange,
    ServiceSearchResultsAggregations,
    ServiceSortableField,
)
from argilla.server.services.tasks.commons import (
    ServiceBaseAnnotation,
    ServiceBaseRecord,
    ServiceBaseRecordInputs,
)


class SortableField(ServiceSortableField):
    pass


class BaseAnnotation(ServiceBaseAnnotation):
    pass


Annotation = TypeVar("Annotation", bound=BaseAnnotation)


class BaseRecordInputs(ServiceBaseRecordInputs[Annotation], Generic[Annotation]):
    def extended_fields(self) -> Dict[str, Any]:
        return {}


class BaseRecord(ServiceBaseRecord[Annotation], Generic[Annotation]):
    pass


class ScoreRange(ServiceQueryRange):
    pass


_Record = TypeVar("_Record", bound=BaseRecord)


class BulkResponse(BaseModel):
    dataset: str
    processed: int
    failed: int = 0


class BaseSearchResults(GenericModel, Generic[_Record, ServiceSearchResultsAggregations]):
    total: int = 0
    records: List[_Record] = Field(default_factory=list)
    aggregations: ServiceSearchResultsAggregations = None
